import os
import warnings

import astropy.units as u
import numpy as np
from astropy.time import Time

from sora.config import input_tests
from sora.config.decorators import deprecated_alias
from .utils import calc_fresnel

warnings.simplefilter('always', UserWarning)


class LightCurve:
    """Defines a Light Curve.

    Parameters
    ----------
    name : `str`
        The name of the LightCurve. Each time an LightCurve object is defined
        the name must be different.

    tref : `astropy.time.Time`, `str`, `float`
        Instant of reference.

        Format: `Julian Date`, string in ISO format or Time object.
        Required only if LightCurve have input fluxes and given time is
        not in Julian Date.

    central_bandpass : `int`, `float`, otpional, default=0.7
        The center band pass of the detector used in observation. Value in microns.

    delta_bandpass : `int`, `float`, optional, default=0.3
        The band pass width of the detector used in observation. Value in microns.

    exptime : `int`, `float`
        The exposure time of the observation, in seconds.
        *NOT* required in cases *2*, *3* and *4* below.
        *Required* in case *1* below.

    **kwargs: `int`, `float`
        Object velocity, distance, and star diameter.

        Note
        ----
        vel : `int`, `float`
            Velocity in km/s.

        dist : `int`, `float`
            Object distance in AU.

        d_star : `float`
            Star diameter, in km.


    Warning
    -------
    Input data must be one of the 4 options below:

    1) Input data from file with time and flux
        `file (str)`: a file with the time and flux. A third column with the error in
        flux can also be given.

        `usecols (int, tuple, array)`: Which columns to read, with the first being the
        time, the seconds the flux and third the flux error (optional).

        **Example:**

        >>> LightCurve(name, file, exptime) # dflux can also be given


    2) Input data when file is not given:
        `time`: time must be a list of times, in seconds from tref, or Julian Date, or
        a Time object.

        `flux`: flux must be a list of fluxes. It must have the same lenght as time.

        `dflux`: if file not given, dflux must be a list of fluxes errors. It must
        have the same lenght as time. (not required)

        **Example:**

        >>> LightCurve(name, flux, time, exptime) # dflux can also be given


    Cases for when `time` and `flux` are not given.


    3) Input for a positive occultation:
        `immersion`: The instant of immersion.

        `emersion`: The instant of emersion.

        `immersion_err`: Immersion time uncertainty, in seconds.

        `emersion_err`: Emersion time uncertainty, in seconds.

        **Example:**

        >>> LightCurve(name, immersion, immersion_err, emersion, emersion_err)


    4) Input for a negative occultation:
        `initial_time`: The initial time of observation.

        `end_time`: The end time of observation.

        **Example:**

        >>> LightCurve(name, initial_time, end_time)

    """

    @deprecated_alias(lambda_0='central_bandpass', delta_lambda='delta_bandpass')  # remove this line for v1.0
    def __init__(self, name='', **kwargs):

        allowed_kwargs = ['emersion', 'emersion_err', 'immersion', 'immersion_err', 'initial_time', 'end_time',
                          'file', 'time', 'flux', 'exptime', 'central_bandpass', 'delta_bandpass', 'tref', 'dflux',
                          'skiprows', 'usecols', 'dist', 'vel', 'd_star']
        input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)
        input_done = False
        self.dflux = None
        self._name = name
        self.flux = None
        self.time_model = None
        if 'tref' in kwargs:
            self.tref = kwargs['tref']
        if 'immersion' in kwargs:
            self.immersion = kwargs['immersion']
            self.immersion_err = kwargs.get('immersion_err', 0.0)
            if self.immersion_err < 0:
                warnings.warn("Immersion Error must be positive. Using absolute value.")
                self.immersion_err = np.absolute(self.immersion_err)
            input_done = True
        if 'emersion' in kwargs:
            self.emersion = kwargs['emersion']
            self.emersion_err = kwargs.get('emersion_err', 0.0)
            if self.emersion_err < 0:
                warnings.warn("Emersion Error must be positive. Using absolute value.")
                self.emersion_err = np.absolute(self.emersion_err)
            try:
                if self.emersion <= self.immersion:
                    raise ValueError("emersion time must be greater than immersion time")
            except AttributeError:
                pass
            input_done = True
        if 'initial_time' in kwargs and 'end_time' in kwargs:
            self.initial_time = kwargs['initial_time']
            self.end_time = kwargs['end_time']
            if self.end_time <= self.initial_time:
                raise ValueError('end_time must be greater than initial_time')
            input_done = True
        if not input_done:
            try:
                self.set_flux(**kwargs)
            except:
                raise ValueError('No allowed input conditions satisfied. Please refer to the tutorial.')
        self.set_filter(central_bandpass=kwargs.get('central_bandpass', 0.70),
                        delta_bandpass=kwargs.get('delta_bandpass', 0.30))
        self.dt = 0.0

    @property
    def fresnel_scale(self):
        lamb = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist = self.dist*u.au.to('km')
        fresnel_scale_1 = calc_fresnel(dist, lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist, lamb+dlamb/2.0)
        fresnel_scale = (fresnel_scale_1 + fresnel_scale_2)/2.0
        return fresnel_scale

    @property
    def central_bandpass(self):
        return self.lambda_0

    @property
    def delta_bandpass(self):
        return self.delta_lambda

    @property
    def name(self):
        return self._name

    @property
    def tref(self):
        if hasattr(self, '_tref'):
            return self._tref
        else:
            raise AttributeError("'LightCurve' object has no attribute 'tref'")

    @tref.setter
    def tref(self, value):
        if type(value) in [int, float]:
            self.tref = Time(value, format='jd')
        else:
            try:
                self._tref = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by tref'.format(value))

    @property
    def immersion(self):
        if hasattr(self, '_immersion'):
            return self._immersion + self.dt*u.s
        else:
            raise AttributeError('The immersion time was not fitted or instantiated.')

    @immersion.setter
    def immersion(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.immersion = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.immersion = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._immersion = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by immersion'.format(value))

    @property
    def emersion(self):
        if hasattr(self, '_emersion'):
            return self._emersion + self.dt*u.s
        else:
            raise AttributeError('The emersion time was not fitted or instanciated.')

    @emersion.setter
    def emersion(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.emersion = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.emersion = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._emersion = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by emersion'.format(value))

    @property
    def initial_time(self):
        if hasattr(self, '_initial_time'):
            return self._initial_time
        else:
            raise AttributeError("'LightCurve' object has no attribute 'initial_time'")

    @initial_time.setter
    def initial_time(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.initial_time = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.initial_time = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._initial_time = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by initial_time'.format(value))

    @property
    def end_time(self):
        if hasattr(self, '_end_time'):
            return self._end_time
        else:
            raise AttributeError("'LightCurve' object has no attribute 'end_time'")

    @end_time.setter
    def end_time(self, value):
        if type(value) in [int, float]:
            if value > 2400000:
                self.end_time = Time(value, format='jd')
            elif hasattr(self, 'tref'):
                self.end_time = self.tref + value*u.s
            else:
                raise ValueError('{} can not be set without a reference time'.format(value))
        else:
            try:
                self._end_time = Time(value)
            except ValueError:
                raise ValueError('{} is not a valid time format accepted by end_time'.format(value))

    @property
    def time_mean(self):
        if hasattr(self, '_immersion') and hasattr(self, '_emersion'):
            return Time((self.immersion.jd + self.emersion.jd)/2, format='jd')
        else:
            return Time((self.initial_time.jd + self.end_time.jd)/2, format='jd')

    @property
    def time(self):
        try:
            return (self._time - self.tref).sec
        except:
            raise AttributeError("'LightCurve' object has no attribute 'time'")

    def set_flux(self, **kwargs):
        """Sets the flux for the LightCurve.

        Parameters
        ----------
        exptime : `int`, `float`, required
            The exposure time of the observation, in seconds.

        file : `str`
            A file with the time and flux in the first and second columns,
            respectively. A third column with error in flux can also be given.

        time
            If file not given, time must be a list of times, in seconds from `tref`,
            or `Julian Date`, or a `Time object`.

        flux
            If file not given, flux must be a list of fluxes. It must have the
            same lenght as time.

        dflux
            If file not given, dflux must be a list of fluxes errors. It must
            have the same lenght as time.

        tref : `astropy.time.Time`, `str`, `float`
            Instant of reference. It can be in `Julian Date`, string in ISO
            format or `Time object`.

        usecols : `int`, `tuple`, array, optional
            Which columns to read, with the first being the time, the seconds
            the flux and third the flux error.

        **kwargs : `int`, `float`
            Object velocity, object distance, star diameter.

            Note
            ----
            vel : `int`, `float`
                Velocity in km/s.

            dist : `int`, `float`:
                Object distance in AU.

            d_star : `float`
                Star diameter, in km.
        """
        from .utils import read_lc_file

        input_done = False
        usecols = None
        if 'usecols' in kwargs:
            usecols = kwargs['usecols']
        skiprows = 0
        if 'skiprows' in kwargs:
            skiprows = int(kwargs['skiprows'])
        if 'file' in kwargs:
            try:
                lc_data = read_lc_file(kwargs['file'], usecols=usecols, skiprows=skiprows)
                if len(lc_data) == 2:
                    time, self.flux = lc_data
                elif len(lc_data) == 3:
                    time, self.flux, self.dflux = lc_data
            except:
                pass
            if hasattr(self, 'flux'):
                self.flux_obs = self.flux
            if not hasattr(self, 'flux_obs'):
                raise ValueError('Input file must have 2 or 3 columns')
            input_done = True
        if 'time' in kwargs and 'flux' in kwargs:
            if input_done:
                raise ValueError('Only one type of input can be given. Please refer to the tutorial.')
            self.flux = kwargs['flux']
            time = kwargs['time']
            if len(self.flux) != len(time):
                raise ValueError('time and flux must have the same length')
            if 'dflux' in kwargs:
                self.dflux = kwargs['dflux']
                if len(self.flux) != len(self.dflux):
                    raise ValueError('dflux must have the same length as flux and time')
            input_done = True
        if 'exptime' not in kwargs:
            raise ValueError('exptime not defined')
        if kwargs['exptime'] <= 0:
            raise ValueError('Exposure time can not be zero or negative')
        else:
            self.exptime = kwargs['exptime']
        if 'vel' in kwargs:
            self.set_vel(vel=kwargs['vel'])
        if 'dist' in kwargs:
            self.set_dist(dist=kwargs['dist'])
        if 'd_star' in kwargs:
            self.set_star_diam(d_star=kwargs['d_star'])
        if 'tref' in kwargs:
            self.tref = kwargs['tref']
        if 'time' in locals():
            if type(time) == Time:
                if not hasattr(self, 'tref'):
                    self.tref = Time(time[0].iso.split(' ')[0] + ' 00:00:00.000')
            elif all(time > 2400000):
                time = Time(time, format='jd')
                if not hasattr(self, 'tref'):
                    self.tref = Time(time[0].iso.split(' ')[0] + ' 00:00:00.000')
            elif not hasattr(self, 'tref'):
                raise ValueError('tref must be given')
            else:
                time = self.tref + time*u.s
            order = np.argsort(time)
            self._time = time[order]
            self.model = np.ones(len(time))
            self.flux = self.flux[order]
            self.flux_obs = self.flux
            if self.dflux is not None:
                self.dflux = self.dflux[order]
            self.initial_time = np.min(time)
            self.end_time = np.max(time)
            time_diffs = np.diff(self._time[0:]).tolist()
            self.cycle = max(set(time_diffs), key=time_diffs.count).sec
            if self.cycle < self.exptime:
                warnings.warn('Exposure time ({:0.4f} seconds) higher than Cycle time ({:0.4f} seconds)'.
                              format(self.exptime, self.cycle))

    def set_exptime(self, exptime):
        """Sets the light curve exposure time.

        Parameters
        ----------
        exptime : `int`, `float`
            Exposure time, in seconds.
        """
        exptime = u.Quantity(exptime, unit=u.s)
        if not np.isscalar(exptime):
            raise TypeError('Exposure time must be an integer, a float or an Astropy Unit object')
        if exptime.value <= 0:
            raise ValueError('Exposure time can not be zero or negative')
        self.exptime = exptime.value
        try:
            if self.cycle < self.exptime:
                warnings.warn('Exposure time ({:0.4f} seconds) higher than Cycle time ({:0.4f} seconds)'.
                              format(self.exptime, self.cycle))
        except:
            pass

    def set_vel(self, vel):
        """Sets the occultation velocity.

        Parameters
        ----------
        vel : `int`, `float`
            Velocity in km/s.
        """
        vel = u.Quantity(vel, unit=u.km/u.s)
        self.vel = np.absolute(vel.value)

    def set_dist(self, dist):
        """Sets the object distance.

        Parameters
        ----------
        dist : `int`, `float`
            Object distance in AU.
        """
        dist = u.Quantity(dist, unit=u.AU)
        if dist.value < 0:
            warnings.warn("distance cannot be negative. Using absolute value.")
        self.dist = np.absolute(dist.value)

    def set_star_diam(self, d_star):
        """Sets the star diameter.

        Parameters
        ----------
        d_star : `float`
            Star diameter, in km.
        """
        d_star = u.Quantity(d_star, unit=u.km)
        if d_star.value < 0:
            warnings.warn("star diameter cannot be negative. Using absolute value.")
        self.d_star = np.absolute(d_star.value)

    @deprecated_alias(lambda_0='central_bandpass', delta_lambda='delta_bandpass')  # remove this line for v1.0
    def set_filter(self, central_bandpass, delta_bandpass):
        """Sets the filter bandwidth in microns.

        Parameters
        ----------
        central_bandpass : `float`
            Center band in microns.

        delta_bandpass : `float`
            Bandwidth in microns.
        """
        central_bandpass = u.Quantity(central_bandpass, unit=u.micrometer)
        if central_bandpass.value <= 0:
            raise ValueError("central bandpass cannot be negative.")
        self.lambda_0 = central_bandpass.value
        delta_bandpass = u.Quantity(delta_bandpass, unit=u.micrometer)
        if delta_bandpass <= 0:
            raise ValueError("delta bandpass cannot be negative")
        self.delta_lambda = delta_bandpass.value
        if (central_bandpass - delta_bandpass).value <= 0:
            raise ValueError("The given central and delta bandpass give a range ({}, {}) microns. Bandpass cannot be negative. "
                             "Please give appropriate values".format(*(central_bandpass +
                                                                       np.array([-1, 1])*delta_bandpass).value))

    def calc_magnitude_drop(self, mag_star, mag_obj):
        """Determines the magnitude drop of the occultation.

        Parameters
        ----------
        mag_star : `int`, `float`
            Star magnitude.

        mag_obj `int`, `float`
            Object apparent magnitude to the date.

        Returns
        -------
        mag_drop : `float`
            Magnitude drop for the given magnitudes.

        bottom_flux : `float`
            Normalized bottom flux for the given magnitudes.
        """
        from .utils import calc_magnitude_drop
        mag_drop, bottom_flux = calc_magnitude_drop(mag_star, mag_obj)
        self.mag_drop = mag_drop
        self.bottom_flux = bottom_flux

    def normalize(self, poly_deg=None, mask=None, flux_min=0.0, flux_max=1.0, plot=False):
        """Returns the fresnel scale.

        Parameters
        ----------
        poly_deg : `int`
            Degree of the polynomial to be fitted.

        mask : `bool` array
            Which values to be fitted.

        flux_min : `int`, `float`
            Event flux to be set as 0.

        flux_max : `int`, `float`
            Baseline flux to be set as 1.

        plot : `bool`
            If True plot the steps for visual aid.
        """
        from .utils import fit_pol
        import matplotlib.pyplot as plt

        # Create a mask where the polynomial fit will be done
        if not all(self.flux):
            raise ValueError('Normalization is only possible when a LightCurve is instantiated with time and flux.')
        self.reset_flux()
        lc_flux = (self.flux - flux_min)/(flux_max-flux_min)
        if mask is None:
            preliminar_occ = self.occ_detect(maximum_duration=((self.end_time - self.initial_time).value*u.d.to('s'))/3)
            tmax = preliminar_occ['emersion_time']+1.00*preliminar_occ['occultation_duration']
            tmin = preliminar_occ['immersion_time']-1.00*preliminar_occ['occultation_duration']
            chord = preliminar_occ['occultation_duration']
            mask = np.invert((self.time > tmin-(chord/2)) & (self.time < tmax+(chord/2)))
        norm_time = (self.time - self.time.min())/(self.time.max()-self.time.min())
        if poly_deg is not None:
            n = poly_deg
            p, err = fit_pol(norm_time[mask], lc_flux[mask], n)
            flux_poly_model = np.zeros(len(norm_time))
            for ii in np.arange(n+1):
                flux_poly_model = flux_poly_model + p[ii]*(norm_time**(n-ii))
            if plot:
                plt.plot(norm_time[mask], lc_flux[mask], 'k.-')
                plt.plot(norm_time[mask], flux_poly_model[mask], 'r-')
                plt.title('Polynomial degree = {}'.format(n), fontsize=15)
                plt.show()
        else:
            n = 0
            p, err = fit_pol(norm_time[mask], lc_flux[mask], n)
            flux_poly_model = np.zeros(len(norm_time))
            for ii in np.arange(n+1):
                flux_poly_model += p[ii]*(norm_time**(n-ii))
            if plot:
                plt.plot(norm_time[mask], lc_flux[mask], 'k.-')
                plt.plot(norm_time[mask], flux_poly_model[mask], 'r-')
                plt.title('Polynomial degree = {}'.format(n), fontsize=15)
                plt.show()
            for nn in np.arange(1, 10):
                p, err = fit_pol(norm_time[mask], lc_flux[mask], nn)
                flux_poly_model_new = np.zeros(len(norm_time))
                for ii in np.arange(nn+1):
                    flux_poly_model_new += p[ii]*(norm_time**(nn-ii))
                F = np.var(flux_poly_model[mask]-lc_flux[mask])/np.var(flux_poly_model_new[mask]-lc_flux[mask])
                if F > 1.05:
                    flux_poly_model = flux_poly_model_new.copy()
                    n = nn
                    if plot:
                        plt.plot(norm_time[mask], lc_flux[mask], 'k.-')
                        plt.plot(norm_time[mask], flux_poly_model[mask], 'r-')
                        plt.title('Polynomial degree = {}'.format(nn), fontsize=15)
                        plt.show()
                else:
                    print('Normalization using a {} degree polynomial'.format(n))
                    print('There is no improvement with a {} degree polynomial'.format(n+1))
                    break
        self.flux = lc_flux/flux_poly_model
        self.normalizer_flux = flux_poly_model
        self.normalizer_mask = mask

    def reset_flux(self):
        """ Resets flux for original values
        """
        try:
            self.flux = self.flux_obs
        except:
            raise ValueError('Reset is only possible when a LightCurve is instantiated with time and flux.')
        return

    def occ_model(self, immersion_time, emersion_time, opacity, mask, npt_star=12,
                  time_resolution_factor=10, flux_min=0, flux_max=1):
        """Returns the modelled light curve.

        The modelled light curve takes into account the fresnel diffraction, the
        star diameter and the instrumental response.

        Parameters
        ----------
        immersion_time : `int`, `float`
            Immersion time, in seconds.

        emersion_time : `int`, `float`
            Emersion time, in seconds.

        opacity : `int`, `float`
            Opacity. Opaque = 1.0, transparent = 0.0,

        mask : `bool` array
            Mask with True values to be computed.

        npt_star : `int`, default=12
            Number of subdivisions for computing the star size effects.

        time_resolution_factor : `int`, `float`, default: 10*fresnel scale
            Steps for fresnel scale used for modelling the light curve.

        flux_min : `int`, `float`, default=0
            Bottom flux (only object).

        flux_max : `int`, `float`, default=1
            Base flux (object plus star).
        """
        from .utils import bar_fresnel

        # Computing the fresnel scale
        lamb = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist = self.dist*u.au.to('km')
        vel = np.absolute(self.vel)
        time_obs = self.time[mask]
        fresnel_scale_1 = calc_fresnel(dist, lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist, lamb+dlamb/2.0)
        fresnel_scale = (fresnel_scale_1 + fresnel_scale_2)/2.0
        time_resolution = (np.min([fresnel_scale/vel, self.exptime]))/time_resolution_factor

        # Creating a high resolution curve to compute fresnel diffraction, stellar diameter and instrumental integration
        time_model = np.arange(time_obs.min()-5*self.exptime, time_obs.max()+5*self.exptime, time_resolution)

        # Changing X: time (s) to distances in the sky plane (km), considering the tangential velocity (vel in km/s)
        x = time_model*vel
        x01 = immersion_time*vel
        x02 = emersion_time*vel

        # Computing fresnel diffraction for the case where the star size is negligenciable
        flux_fresnel_1 = bar_fresnel(x, x01, x02, fresnel_scale_1, opacity)
        flux_fresnel_2 = bar_fresnel(x, x01, x02, fresnel_scale_2, opacity)
        flux_fresnel = (flux_fresnel_1 + flux_fresnel_2)/2.
        flux_star = flux_fresnel.copy()
        if self.d_star > 0:
            # Computing fresnel diffraction for the case where the star size is not negligenciable
            resolucao = (self.d_star/2)/npt_star
            flux_star_1 = np.zeros(len(time_model))
            flux_star_2 = np.zeros(len(time_model))
            # Computing stellar diameter only near the immersion or emersion times
            star_diam = (np.absolute(x - x01) < 3*self.d_star) + (np.absolute(x - x02) < 3*self.d_star)
            p = np.arange(-npt_star, npt_star)*resolucao
            coeff = np.sqrt(np.absolute((self.d_star/2)**2 - p**2))
            for ii in np.where(star_diam == True)[0]:
                xx = x[ii] + p
                flux1 = bar_fresnel(xx, x01, x02, fresnel_scale_1, opacity)
                flux2 = bar_fresnel(xx, x01, x02, fresnel_scale_2, opacity)
                flux_star_1[ii] = np.sum(coeff*flux1)/coeff.sum()
                flux_star_2[ii] = np.sum(coeff*flux2)/coeff.sum()
                flux_star[ii] = (flux_star_1[ii] + flux_star_2[ii])/2.
        flux_inst = np.zeros(len(time_obs))
        for i in range(len(time_obs)):
            event_model = (time_model > time_obs[i]-self.exptime/2.) & (time_model < time_obs[i]+self.exptime/2.)
            flux_inst[i] = (flux_star[event_model]).mean()
        self.model[mask] = flux_inst*(flux_max - flux_min) + flux_min
        self.time_model = time_model
        self.model_star = flux_star*(flux_max - flux_min) + flux_min
        self.model_fresnel = flux_fresnel*(flux_max - flux_min) + flux_min
        ev_model = (time_model > immersion_time) & (time_model < emersion_time)
        flux_box = np.ones(len(time_model))
        flux_box[ev_model] = (1-opacity)**2
        flux_box = flux_box*(flux_max - flux_min) + flux_min
        self.model_geometric = flux_box
        self.baseflux = flux_max
        self.bottomflux = flux_min

    def occ_lcfit(self, **kwargs):
        """Monte Carlo chi square fit for occultations lightcurve.

        Parameters
        ----------
        tmin : `int`, `float`
            Minimum time to consider in the fit procedure, in seconds.

        tmax : `int`, `float`
            Maximum time to consider in the fit procedure, in seconds.

        flux_min : `int`, `float`, default=0
            Bottom flux (only object).

        flux_max :`int`, `float`, default=1
            Base flux (object plus star).

        immersion_time : `int`, `float`
            Initial guess for immersion time, in seconds.

        emersion_time : `int`, `float`
            Initial guess for emersion time, in seconds.

        opacity : `int`, `float`, default=1
            Initial guess for opacity. Opaque = 1, Transparent = 0.

        delta_t : `int`, `float`
            Interval to fit immersion or emersion time.

        dopacity : `int`, `float`, default=0
            Interval to fit opacity.

        sigma : `int`, `float`, `array`, 'auto'
            Fluxes errors. If None it will use the `self.dflux`. If 'auto' it
            will calculate using the region outside the event.

        loop : `int`, default=10000
            Number of tests to be done.

        sigma_result : `int`, `float`
            Sigma value to be considered as result.

        Returns
        -------
        chi2 : `sora.extra.ChiSquare`
            ChiSquare object.
        """
        from sora.config.visuals import progressbar
        from sora.extra import ChiSquare

        allowed_kwargs = ['tmin', 'tmax', 'flux_min', 'flux_max', 'immersion_time', 'emersion_time', 'opacity',
                          'delta_t', 'dopacity', 'sigma', 'loop', 'sigma_result']
        input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)

        if not hasattr(self, 'flux'):
            raise ValueError('Fit curve is only possible when a LightCurve is instantiated with time and flux.')

        preliminar_occ = self.occ_detect()

        delta_t = 2*self.cycle
        loop = kwargs.get('loop', 10000)
        tmax = self.time.max()
        tmin = self.time.min()
        immersion_time = tmin - self.exptime
        do_immersion = False
        emersion_time = tmax + self.exptime
        do_emersion = False
        opacity = kwargs.get('opacity', 1.0)
        delta_opacity = kwargs.get('dopacity', 0.0)
        do_opacity = 'dopacity' in kwargs
        if ('immersion_time' not in kwargs) and ('emersion_time' not in kwargs):
            immersion_time = preliminar_occ['immersion_time']
            do_immersion = True
            emersion_time = preliminar_occ['emersion_time']
            do_emersion = True
            delta_t = 5*preliminar_occ['time_err']
            tmax = emersion_time+2*preliminar_occ['occultation_duration']
            tmin = immersion_time-2*preliminar_occ['occultation_duration']
            if 2*preliminar_occ['occultation_duration'] < 10*self.cycle:
                tmax = emersion_time + 10*self.cycle
                tmin = immersion_time - 10*self.cycle
        tmax = kwargs.get('tmax', tmax)
        tmin = kwargs.get('tmin', tmin)
        delta_t = kwargs.get('delta_t', delta_t)
        if 'immersion_time' in kwargs:
            immersion_time = kwargs['immersion_time']
            do_immersion = True
        t_i = immersion_time + delta_t*(2*np.random.random(loop) - 1)
        if 'emersion_time' in kwargs:
            emersion_time = kwargs['emersion_time']
            do_emersion = True
        t_e = emersion_time + delta_t*(2*np.random.random(loop) - 1)
        mask = (self.time >= tmin) & (self.time <= tmax)
        if 'sigma' not in kwargs:
            if self.dflux is not None:
                sigma = self.dflux
            else:
                sigma = 'auto'
        else:
            if type(kwargs['sigma']) in [float, int]:
                sigma = np.repeat(kwargs['sigma'], len(self.flux))
            elif kwargs['sigma'] is None:
                sigma = self.dflux
            else:
                sigma = kwargs['sigma']
        if type(sigma) is str and sigma == 'auto':
            mask_sigma = (((self.time >= tmin) & (self.time < immersion_time - self.exptime)) +
                          ((self.time > emersion_time + self.exptime) & (self.time <= tmax)))
            sigma = np.repeat(self.flux[mask_sigma].std(ddof=1), len(self.flux))
        opas = opacity + delta_opacity*(2*np.random.random(loop) - 1)
        opas[opas > 1.], opas[opas < 0.] = 1.0, 0.0
        flux_min = kwargs.get('flux_min', 1 - preliminar_occ['depth'])
        flux_max = kwargs.get('flux_max', preliminar_occ['baseline'])
        sigma_result = kwargs.get('sigma_result', 1)

        tflag = np.zeros(loop)
        tflag[t_i > t_e] = t_i[t_i > t_e]
        t_i[t_i > t_e] = t_e[t_i > t_e]
        t_e[t_i > t_e] = tflag[t_i > t_e]
        chi2 = 999999*np.ones(loop)
        for i in progressbar(range(loop), 'LightCurve fit:'):
            model_test = self.__occ_model(t_i[i], t_e[i], opas[i], mask, flux_min=flux_min, flux_max=flux_max)
            chi2[i] = np.sum(((self.flux[mask] - model_test)**2)/(sigma[mask]**2))
        kkwargs = {}
        if do_immersion:
            kkwargs['immersion'] = t_i
        if do_emersion:
            kkwargs['emersion'] = t_e
        if do_opacity:
            kkwargs['opacity'] = opas
        chisquare = ChiSquare(chi2, len(self.flux[mask]), **kkwargs)
        result_sigma = chisquare.get_nsigma(sigma=sigma_result)
        if 'immersion' in result_sigma:
            self._immersion = self.tref + result_sigma['immersion'][0]*u.s
            self.immersion_err = result_sigma['immersion'][1]
            immersion_time = result_sigma['immersion'][0]
        else:
            try:
                immersion_time = (self._immersion.jd - self.tref.jd)*u.d.to('s')
            except:
                pass
        if 'emersion' in result_sigma:
            self._emersion = self.tref + result_sigma['emersion'][0]*u.s
            self.emersion_err = result_sigma['emersion'][1]
            emersion_time = result_sigma['emersion'][0]
        else:
            try:
                emersion_time = (self._emersion.jd - self.tref.jd)*u.d.to('s')
            except:
                pass
        if 'opacity' in result_sigma:
            opacity = result_sigma['opacity'][0]
        # Run occ_model() to save best parameters in the Object.
        self.occ_model(immersion_time, emersion_time, opacity, np.repeat(True, len(self.flux)),
                       flux_min=flux_min, flux_max=flux_max)
        self.lc_sigma = sigma
        self.chisquare = chisquare
        self.opacity = opacity
        return chisquare

    def plot_lc(self, ax=None):
        """ Plots the light curve
        """
        import matplotlib.pyplot as plt

        if not any(self.flux):
            raise ValueError('Plotting the light curve is only possible when the '
                             'Object LightCurve is instantiated with time and flux')
        ax = ax or plt.gca()
        ax.plot(self.time, self.flux, 'k.-', label='Obs.', zorder=0)
        if any(self.model):
            ax.plot(self.time, self.model, 'r-', label='Model', zorder=2)
            ax.scatter(self.time, self.model, s=50, facecolors='none', edgecolors='r', zorder=3)
        ax.set_xlabel('Time [seconds]', fontsize=20)
        ax.set_ylabel('Relative Flux', fontsize=20)
        ax.legend()

    def plot_model(self, ax=None):
        """ Plots the modelled light curve
        """
        import matplotlib.pyplot as plt

        if not all(self.time_model):
            raise ValueError('Plotting the model light curve is only possible after the model '
                             '[LightCurve.occ_model()] or the fit [LightCurve.occ_lcfit()]')
        ax = ax or plt.gca()
        ax.plot(self.time_model, self.model_geometric, 'c-', label='Geometric', zorder=1)
        ax.plot(self.time_model, self.model_fresnel, 'b-', label='Fresnel', zorder=1)
        ax.plot(self.time_model, self.model_star, 'g-', label='Star diam.', zorder=1)
        ax.set_xlabel('Time [seconds]', fontsize=20)
        ax.set_ylabel('Relative Flux', fontsize=20)
        ax.legend()

    def to_log(self, namefile=None):
        """Saves the light curve log to a file.

        Parameters
        ----------
        namefile : `str`
            Filename to save the log.
        """
        if namefile is None:
            namefile = self.name.replace(' ', '_')+'.log'
        f = open(namefile, 'w')
        f.write(self.__str__())
        f.close()

    def to_file(self, namefile=None):
        """Saves the light curve to a file.

        Parameters
        ----------
        namefile : `str`
            Filename to save the data.
        """
        # Observational data
        if namefile is None:
            folder = ''
            file = self.name.replace(' ', '_')+'.dat'
        else:
            folder = os.path.dirname(namefile)
            file = os.path.basename(namefile)
        data = np.array([(self.time*u.s + self.tref).jd, self.time, self.flux, self.model, self.flux-self.model])
        colunm_names = ['Time JD', 'Time relative to {} UTC in seconds'.format(self.tref.iso),
                        'Observational Flux', 'Modelled Flux', 'Residual O-C']
        np.savetxt(os.path.join(folder, file), data.T, fmt='%11.8f')
        f = open(os.path.join(folder, file) + '.label', 'w')
        for i, name in enumerate(colunm_names):
            f.write('Column {}: {}\n'.format(i+1, name))
        f.close()
        # Complete Model
        if all(self.time_model):
            data_model = np.array([(self.time_model*u.s + self.tref).jd, self.time_model, self.model_geometric,
                                   self.model_fresnel, self.model_star])
            colunm_names_model = ['Model time JD', 'Model time relative to {} UTC in seconds'.format(self.tref.iso),
                                  'Geometric Model', 'Model with Fresnel diffraction', 'Model with star diameter']
            np.savetxt(os.path.join(folder, 'model_'+file), data_model.T, fmt='%11.8f')
            f = open(os.path.join(folder, 'model_'+file)+'.label', 'w')
            for i, name in enumerate(colunm_names_model):
                f.write('Column {}: {}\n'.format(i+1, name))
            f.close()

    def occ_detect(self, maximum_duration=None, dur_step=None, snr_limit=None,
                   n_detections=None, plot=False):
        """Detects automatically the occultation event in the light curve.

        Detects a 'square well' shaped transit. All parameters are optional.

        Parameters
        ----------
        maximum_duration : `float`, default: light curve time span
            Maximum duration of the occultation event.

        dur_step : `float`, default: 1/2 of sampling rate
            Step size to sweep occultation duration event.

        snr_limit : `float`, default=None
            Minimum occultation SNR.

        n_detections : `int`, default=1
            Number of detections regardless the SNR. `n_detections` is
            superseded by `snr_limit`.

        plot : `bool`
            True if output plots are desired.


        Returns
        -------
        OrderedDict : `dict`
            An ordered dictionary of :attr:`name`::attr:`value` pairs for each
            parameter.

        Examples
        --------
        >>> lc = LightCurve(time=time, flux=flux, exptime=0.0, name='lc_example')
        >>> params = lc.occ_detect()
        >>> params
        {'rank': 1,
        'occultation_duration': 40.1384063065052,
        'central_time': 7916.773870512843,
        'immersion_time': 7896.7046673595905,
        'emersion_time': 7936.843073666096,
        'time_err': 0.05011036992073059,
        'depth': 0.8663887801707082,
        'depth_err': 0.10986223384336465,
        'baseline': 0.9110181732552853,
        'baseline_err': 0.19045768512595365,
        'snr': 7.886138392251848,
        'occ_mask': array([False, False, False, ..., False, False, False])}
        """
        from .occdetect import occ_detect
        occ = occ_detect(self.flux, self.dflux, self.time, self.cycle, maximum_duration=maximum_duration,
                         dur_step=dur_step, snr_limit=snr_limit, n_detections=n_detections, plot=plot)
        return occ

    def __occ_model(self, immersion_time, emersion_time, opacity, mask, npt_star=12,
                    time_resolution_factor=10, flux_min=0.0, flux_max=1.0):
        """Private function. Returns the modelled light curve.

        Returns the modelled light curve considering fresnel diffraction, star
        diameter and instrumental response, intended for fitting inside the
        `self.occ_lcfit()`.

        Parameters
        ----------
        immersion_time : `int`, `float`
            Immersion time, in seconds.

        emersion_time : `int`, `float`
            Emersion time, in seconds.

        opacity `int`, `float`
            Opacity. Opaque = 1, Transparent = 0.

        mask : `bool` array
            Mask with True values to be computed.

        npt_star : `int`, default=12
            Number of subdivisions for computing the star size effects.

        time_resolution_factor : `int`, `float`, default=10*fresnel scale
            Steps for fresnel scale used for modelling the light curve.

        flux_min : `int`, `float`, default=0
            Bottom flux (only object).

        flux_max : `int`, `float`, default=1
            Base flux (object plus star).

        Returns
        -------
        flux_inst : array
            Modelled Instrumental light flux.
        """
        from .utils import bar_fresnel

        # Computing the fresnel scale
        lamb = self.lambda_0*u.micrometer.to('km')
        dlamb = self.delta_lambda*u.micrometer.to('km')
        dist = self.dist*u.au.to('km')
        vel = np.absolute(self.vel)
        time_obs = self.time[mask]
        fresnel_scale_1 = calc_fresnel(dist, lamb-dlamb/2.0)
        fresnel_scale_2 = calc_fresnel(dist, lamb+dlamb/2.0)
        fresnel_scale = (fresnel_scale_1 + fresnel_scale_2)/2.0
        time_resolution = (np.min([fresnel_scale/vel, self.exptime]))/time_resolution_factor
        self.model_resolution = time_resolution

        # Creating a high resolution curve to compute fresnel diffraction, stellar diameter and instrumental integration
        time_model = np.arange(time_obs.min()-5*self.exptime, time_obs.max()+5*self.exptime, time_resolution)

        # Changing X: time (s) to distances in the sky plane (km), considering the tangential velocity (vel in km/s)
        x = time_model*vel
        x01 = immersion_time*vel
        x02 = emersion_time*vel

        # Computing fresnel diffraction for the case where the star size is negligenciable
        flux_fresnel_1 = bar_fresnel(x, x01, x02, fresnel_scale_1, opacity)
        flux_fresnel_2 = bar_fresnel(x, x01, x02, fresnel_scale_2, opacity)
        flux_fresnel = (flux_fresnel_1 + flux_fresnel_2)/2.
        flux_star = flux_fresnel.copy()
        if self.d_star > 0:
            # Computing fresnel diffraction for the case where the star size is not negligenciable
            resolucao = (self.d_star/2)/npt_star
            flux_star_1 = np.zeros(len(time_model))
            flux_star_2 = np.zeros(len(time_model))
            # Computing stellar diameter only near the immersion or emersion times
            star_diam = (np.absolute(x - x01) < 3*self.d_star) + (np.absolute(x - x02) < 3*self.d_star)
            p = np.arange(-npt_star, npt_star)*resolucao
            coeff = np.sqrt(np.absolute((self.d_star/2)**2 - p**2))
            for ii in np.where(star_diam == True)[0]:
                xx = x[ii] + p
                flux1 = bar_fresnel(xx, x01, x02, fresnel_scale_1, opacity)
                flux2 = bar_fresnel(xx, x01, x02, fresnel_scale_2, opacity)
                flux_star_1[ii] = np.sum(coeff*flux1)/coeff.sum()
                flux_star_2[ii] = np.sum(coeff*flux2)/coeff.sum()
                flux_star[ii] = (flux_star_1[ii] + flux_star_2[ii])/2.
        flux_inst = np.zeros(len(time_obs))
        for i in range(len(time_obs)):
            event_model = (time_model > time_obs[i]-self.exptime/2.) & (time_model < time_obs[i]+self.exptime/2.)
            flux_inst[i] = (flux_star[event_model]).mean()
        return flux_inst*(flux_max - flux_min) + flux_min

    def __str__(self):
        """ String representation of the LightCurve Object
        """
        output = 'Light curve name: {}\n'.format(self.name)
        try:
            output += ('Initial time: {} UTC\n'
                       'End time:     {} UTC\n'
                       'Duration:     {:.3f} minutes\n'.format(
                           self.initial_time.iso, self.end_time.iso,
                           (self.end_time - self.initial_time).value*u.d.to('min'))
                       )
        except:
            pass
        output += 'Time offset:  {:.3f} seconds\n\n'.format(self.dt)
        try:
            output += 'Exposure time:    {:.4f} seconds\n'.format(self.exptime)
            output += 'Cycle time:       {:.4f} seconds\n'.format(self.cycle)
            output += 'Num. data points: {}\n\n'.format(len(self.time))
        except:
            output += 'Object LightCurve was not instantiated with time and flux.\n\n'
        try:
            output += ('Bandpass:             {:.3f} +/- {:.3f} microns\n'
                       'Object Distance:      {:.2f} AU\n'
                       'Used shadow velocity: {:.3f} km/s\n'
                       'Fresnel scale:        {:.3f} seconds or {:.2f} km\n'
                       'Stellar size effect:  {:.3f} seconds or {:.2f} km\n'.format(
                           self.lambda_0, self.delta_lambda, self.dist, self.vel,
                           self.fresnel_scale/self.vel, self.fresnel_scale,
                           self.d_star/self.vel, self.d_star)
                       )
        except:
            output += '\nThere is no occultation associated with this light curve.\n'
        try:
            output += ('Inst. response:       {:.3f} seconds or {:.2f} km\n'
                       'Dead time effect:     {:.3f} seconds or {:.2f} km\n'
                       'Model resolution:     {:.3f} seconds or {:.2f} km\n'
                       'Modelled baseflux:    {:.3f}\n'
                       'Modelled bottomflux:  {:.3f}\n'
                       'Light curve sigma:    {:.3f}\n\n'.format(
                           self.exptime, self.exptime*self.vel, self.cycle-self.exptime,
                           (self.cycle-self.exptime)*self.vel, self.model_resolution,
                           self.model_resolution*self.vel, self.baseflux, self.bottomflux,
                           self.lc_sigma.mean())
                       )
        except:
            output += '\nObject LightCurve model was not fitted.\n\n'
        try:
            output += ('Immersion time: {} UTC +/- {:.3f} seconds\n'
                       'Emersion time:  {} UTC +/- {:.3f} seconds\n\n'.format(
                           self.immersion.iso, self.immersion_err,
                           self.emersion.iso, self.emersion_err)
                       )
        except:
            output += 'Immersion and emersion times were not fitted or instantiated.\n\n'

        try:
            output += 'Monte Carlo chi square fit.\n\n' + self.chisquare.__str__() + '\n'
        except:
            pass
        return output



import glob
import os
import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord, get_sun, get_moon
from astropy.table import Table, Row, Column
from astropy.time import Time

from sora.config import input_tests

__all__ = ['PredictionTable']


class PredictRow(Row):
    """An Astropy Row object modified for Prediction purposes.
    """

    def plot_occ_map(self, **kwargs):
        """
        Parameters
        ----------
        radius : `int`, `float`
            The radius of the shadow. If not given it uses saved value.

        nameimg : `str`
            Change the name of the image saved.

        path : `str`
            Path to a directory where to save map.

        resolution : `int`, default=2
        Cartopy feature resolution.\n
        - ``1`` means a resolution of "10m";\n
        - ``2`` a resolution of "50m";\n
        - ``3`` a resolution of "100m".

        states : `bool`
            If True, plots the states borders of the countries. The states
            of some countries will only be shown depending on the resolution.

        zoom : `int`, `float`
            Zooms in or out of the map.

        centermap_geo : `list`, default=None
            Center the map given coordinates in longitude and latitude. It must be
            a list with two numbers.

        centermap_delta : `list`, default=None
            Displace the center of the map given displacement in X and Y, in km.
            It must be a list with two numbers.

        centerproj : `list`
            Rotates the Earth to show occultation with the center projected at a
            given longitude and latitude. It must be a list with two numbers.

        labels : `bool`, default=True
            Plots text above and below the map with the occultation parameters.

        meridians : `int`, default=30
            Plots lines representing the meridians for given interval, in degrees.

        parallels : `int`, default=30
            Plots lines representing the parallels for given interval, in degrees.

        sites : `dict`
            Plots site positions in map. It must be a python dictionary where the
            key is  the `name` of the site, and the value is a list with `longitude`,
            `latitude`, `delta_x`, `delta_y` and `color`. `delta_x` and `delta_y`
            are displacement, in km, from the point position of the site in the map
            and the `name`. `color` is the color of the point.

        site_name : `bool`
            If True, it prints the name of the sites given, else it plots only the points.

        countries : `dict`
            Plots the names of countries. It must be a python dictionary where the
            key is the name of the country and the value is a list with longitude
            and latitude of the lower left part of the text.

        offset : `list`
            Applies an offset to the ephemeris, calculating new CA and instant of
            CA. It is a pair of `delta_RA*cosDEC` and `delta_DEC`.

        mapstyle : `int`, default=1
            Define the color style of the map. ``'1'`` is the default black
            and white scale. ``'2'`` is a colored map.

        error : `int`, `float`
            Ephemeris error in mas. It plots a dashed line representing radius + error.

        ercolor : `str`
            Changes the color of the lines of the error bar.

        ring : `int`, `float`
            Plots a dashed line representing the location of a ring. It is given
            in km, from the center.

        rncolor : `str`
            Changes the color of ring lines.

        atm : `int`, `float`
            Plots a dashed line representing the location of an atmosphere. It is
            given in km, from the center.

        atcolor : `str`
            Changes the color of atm lines.

        chord_delta : `list`
            List with distances from center to plot chords.

        chord_geo : `2d-list`
            List with pairs of coordinates to plot chords.

        chcolor : `str`, default='grey'
            Color of the line of the chords.

        heights : `list`
            It plots a circular dashed line showing the locations where the observer
            would observe the occultation at a given height above the horizons.
            This must be a list.

        hcolor : `str`
            Changes the color of the height lines.

        mapsize : `list`, default= [46.0, 38.0]
            The size of figure, in cm. It must be a list with two values.

        cpoints : `int`, `float`, default=60
            Interval for the small points marking the center of shadow, in seconds.

        ptcolor : `str`
            Change the color of the center points.

        alpha : `float`, default=0.2
            The transparency of the night shade, where 0.0 is full transparency and
            1.0 is full black.

        fmt : `str`, default:'png'
            The format to save the image. It is parsed directly by `matplotlib.pyplot`.

        dpi : `int`, default=100
            Resolution in "dots per inch". It defines the quality of the image.

        lncolor : `str`
            Changes the color of the line that represents the limits of the shadow
            over Earth.

        outcolor :`str`
            Changes the color of the lines that represents the limits of the shadow
            outside Earth.

        scale : `int`, `float`
            Arbitrary scale for the size of the name of the site.

        cscale : `int`, `float`
            Arbitrary scale for the name of the country.

        sscale : `int`, `float`
            Arbitrary scale for the size of point of the site.

        pscale : `int`, `float`
            Arbitrary scale for the size of the points that represent the center of
            the shadow.

        arrow : `bool`
            If True, it plots the arrow with the occultation direction.


        Note
        ----
        Only one of centermap_geo and centermap_delta can be given.
        """
        from .occmap import plot_occ_map

        radius = kwargs.pop('radius', self.meta['radius'])
        plot_occ_map(self.meta['name'], radius, coord=self['ICRS Star Coord at Epoch'], time=self['Epoch'],
                     ca=float(self['C/A']), pa=float(self['P/A']), vel=float(self['Vel']), dist=float(self['Dist']),
                     mag=float(self['G*']), longi=float(self['long']), **kwargs)


class PredictionTable(Table):
    """ An Astropy Table object modified for Prediction purposes.
    """
    Row = PredictRow

    def __init__(self, *args, **kwargs):
        if 'time' in kwargs:
            values = {}
            time = Time(kwargs['time'])
            time.delta_ut1_utc = 0.0
            time.format = 'iso'
            if time.isscalar:
                time = Time([time])
            values['Epoch'] = Column(time)

            def coord_fmt(value):
                return value.to_string('hmsdms', precision=5, sep=' ')

            coord = SkyCoord(kwargs['coord_star'], unit=(u.hourangle, u.deg))
            values['ICRS Star Coord at Epoch'] = Column(coord, format=coord_fmt)
            try:
                coord_geo = SkyCoord(kwargs['coord_obj'])
            except:
                coord_geo = SkyCoord(kwargs['coord_obj'], unit=(u.hourangle, u.deg))
            values['Geocentric Object Position'] = Column(coord_geo, format=coord_fmt)
            values['C/A'] = Column(kwargs['ca'], format='5.3f', unit='arcsec')
            values['P/A'] = Column(kwargs['pa'], format='6.2f', unit='deg')
            values['Vel'] = Column(kwargs['vel'], format='-6.2f', unit='km/s')
            values['Dist'] = Column(kwargs['dist'], format='7.3f', unit='AU')
            del(kwargs['time'], kwargs['coord_star'], kwargs['coord_obj'], kwargs['ca'],
                kwargs['pa'], kwargs['vel'], kwargs['dist'])
            if 'mag' not in kwargs.keys() and 'mag_20' not in kwargs.keys():
                raise ValueError('User must provide "mag" or "mag_20" parameters')
            if 'mag' in kwargs.keys():
                values['G'] = Column(kwargs['mag'] * u.mag, format='6.3f', unit='mag')
                del kwargs['mag']
            else:
                values['G'] = kwargs['mag_20'] - 2.5*np.log10(np.absolute(values['Vel'])/20.0)
                values['G'].unit = 'mag'
                values['G'].format = '6.3f'
            if 'mag_20' in kwargs.keys():
                values['G*'] = Column(kwargs['mag_20'] * u.mag, format='6.3f', unit='mag')
                del kwargs['mag_20']
            else:
                values['G*'] = values['G'] + 2.5*np.log10(np.absolute(values['Vel'])/20.0)
                values['G*'].unit = 'mag'
                values['G*'].format = '6.3f'
            if 'long' in kwargs.keys():
                values['long'] = Column(kwargs['long'], unit='deg', format='3.0f')
                del kwargs['long']
            else:
                values['long'] = Column((coord.ra - time.sidereal_time('mean', 'greenwich')).wrap_at(360*u.deg),
                                        unit='deg', format='3.0f')
            if 'loct' in kwargs.keys():
                values['loct'] = Column(kwargs['loct'], unit='hh:mm')
                del kwargs['loct']
            else:
                longi = (coord.ra - time.sidereal_time('mean', 'greenwich')).wrap_at(360*u.deg)
                ntime = time + longi.hour*u.hour
                values['loct'] = Column(['{}'.format(t.iso[11:16]) for t in ntime], unit='hh:mm')
            moon_pos = get_moon(time)
            values['M-G-T'] = Column(moon_pos.separation(coord), unit='deg', format='3.0f')
            sun_pos = get_sun(time)
            values['S-G-T'] = Column(sun_pos.separation(coord), unit='deg', format='3.0f')
            catalogue = kwargs.get('meta', {}).get('catalogue', '')
            if 'source' in kwargs.keys():
                values[f'{catalogue} Source ID'] = Column(kwargs['source'], dtype=np.int64)
                del kwargs['source']
            else:
                values[f'{catalogue} Source ID'] = Column(np.repeat('', len(time)))
            super().__init__(values, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    def __itens_by_epoch(self, date):
        """ Gets item list for all occultations that matches the given date

        Parameters
        ----------
        date : `str`
            Date to match

        Returns
        -------
        item : `list`
            The list of occultations that matches the date.
        """
        col = self['Epoch']
        arr = list([i for i, c in enumerate(col) if c.iso.startswith(date)])
        return arr

    def __getitem__(self, item):
        """ The redefinition of __getitem__ allows for selecting prediction based on the ISO date of the event
        """
        if isinstance(item, str) and item not in self.colnames:
            arr = self.__itens_by_epoch(item)
            if len(arr) == 0:
                raise KeyError('No prediction corresponds to time "{}"'.format(item))
            elif len(arr) > 1:
                return self[arr]
            else:
                return self.Row(self, arr[0])
        return super().__getitem__(item)

    @classmethod
    def from_praia(cls, filename, name, **kwargs):
        """Creates a PredictionTable Table reading from a PRAIA table.

        Parameters
        ----------
        filename : `str`
            Path to the PRAIA table file.

        name : `str`
            Name of the Object of the prediction.

        radius : `int`, `float`, optional
            Object radius, in km.
            If not given it's searched in online database.
            When not found online, the default is set to zero.

        Returns
        -------
         : `sora.prediction.PredictionTable`
            A PredictionTable object.
        """
        from sora.body.utils import search_satdb, search_sbdb
        if not os.path.isfile(filename):
            raise IOError('File {} not found'.format(filename))
        input_tests.check_kwargs(kwargs, allowed_kwargs=['radius'])
        try:
            dados = np.loadtxt(
                filename, skiprows=41,
                usecols=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 25, 26, 27, 28, 29),
                dtype={'names': ('dia', 'mes', 'ano', 'hor', 'min', 'sec', 'afh', 'afm', 'afs', 'ded',
                                 'dem', 'des', 'afh2', 'afm2', 'afs2', 'ded2', 'dem2', 'des2', 'ca',
                                 'pa', 'vel', 'delta', 'mR', 'mK', 'long', 'loct', 'ora', 'ode'),
                       'formats': ('S30', 'S30', 'S30', 'S30', 'S30', 'S30', 'S20', 'S20', 'S20',
                                   'S20', 'S20', 'S20', 'S20', 'S20', 'S20', 'S20', 'S20', 'S20',
                                   'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'f4', 'S20', 'f4', 'f4')}, ndmin=1)
        except:
            raise IOError('{} is not in PRAIA format or does not have any occultation'.format(filename))

        f = open(filename, 'r')
        lines = f.readlines()
        f.close()

        max_ca = float(lines[14].split()[-2])*u.arcsec

        # reading coordinates
        coor = np.char.array(dados['afh'], unicode=True)
        for i in ['afm', 'afs', 'ded', 'dem', 'des']:
            coor = np.core.defchararray.add(coor, ' ')
            coor = np.core.defchararray.add(coor, np.char.array(dados[i], unicode=True))
        coord_star = SkyCoord(coor, frame='icrs', unit=(u.hourangle, u.degree))

        coor = np.char.array(dados['afh2'], unicode=True)
        for i in ['afm2', 'afs2', 'ded2', 'dem2', 'des2']:
            coor = np.core.defchararray.add(coor, ' ')
            coor = np.core.defchararray.add(coor, np.char.array(dados[i], unicode=True))
        coord_obj = SkyCoord(coor, frame='icrs', unit=(u.hourangle, u.degree))

        # reading time
        tim = np.char.array(dados['ano'], unicode=True)
        len_iso = ['-', '-', ' ', ':', ':']
        arr = ['mes', 'dia', 'hor', 'min', 'sec']
        for i in np.arange(len(arr)):
            tim = np.core.defchararray.add(tim, len_iso[i])
            tim = np.core.defchararray.add(tim, np.char.array(dados[arr[i]], unicode=True))
        time = Time(np.char.array(tim) + '000')

        # defining parameters
        try:
            data = search_satdb(name.lower())
            radius = data.get('diameter', 0) / 2
        except ValueError:
            try:
                data = search_sbdb(name.lower())
                radius = data.get('diameter', 0)/2
            except ValueError:
                radius = 0
        error_ra, error_dec = 0, 0
        radius = kwargs.get('radius', radius)*u.km
        meta = {'name': name, 'radius': radius, 'max_ca': max_ca, 'ephem': lines[17].split()[-1],
                'error_ra': error_ra*1000, 'error_dec': error_dec*1000}
        return cls(time=time, coord_star=coord_star, coord_obj=coord_obj, ca=dados['ca'], pa=dados['pa'],
                   vel=dados['vel'], mag_20=dados['mR'], dist=dados['delta'],
                   long=dados['long'], loct=dados['loct'], meta=meta)

    def to_praia(self, filename):
        """Writes PredictionTable to PRAIA format.

        Parameters
        ----------
        filename : `str`
            Name of the file to save table.
        """
        from .values import praia_occ_head
        f = open(filename, 'w')
        f.write(praia_occ_head.format(max_ca=self.meta['max_ca'].to(u.arcsec), size=len(self),
                                      ephem=self.meta.get('ephem', 'ephem')))
        for time, coord, coord_geo, ca, pa, vel, dist, mag, mag_20, longi, loct, md, sd, source in self.iterrows():
            dmag = mag_20-mag
            f.write("\n {} {} {}  {}  {}   {}   {:5.3f}  {:6.2f} {:-6.2f} {:5.2f} {:4.1f} "
                    "{:-4.1f} {:-4.1f} {:-4.1f}   {:3.0f}. {}       0.0      0.0 ok g2 0    0    0    0    0".
                    format(time.iso[8:10], time.iso[5:7], time.iso[:4], time.iso[11:21].replace(':', ' '),
                           coord.to_string('hmsdms', precision=4, sep=' '), coord_geo.to_string('hmsdms', precision=4, sep=' '),
                           ca, pa, vel, dist, mag_20, dmag, dmag, dmag, longi, loct))
        f.close()

    def to_ow(self, ow_des, mode='append'):
        """Writes PredictionTable to OccultWatcher feeder update file format.
        Tables will be saved in two files: "tableOccult_update.txt" and "LOG.dat"

        Parameters
        ----------
        ow_des : `str`
            Occult Watcher designation for the object.

        mode : `str`, default='append'
            Use ``'append'`` to append table to already existing file and
            ``'restart'`` to overwrite existing file.
        """
        from .values import ow_occ_head
        modes = {'append': 'a+', 'restart': 'w'}
        if mode not in modes.keys():
            raise ValueError('mode param must be "append" or "restart".')

        f = open('tableOccult_update.txt', modes[mode])
        if self.meta['radius'] == 0.0:
            warnings.warn('radius is 0.0, please update the value manually in "tableOccult_update.txt" '
                          'before submitting the file')
        if (self.meta['error_ra'] == 0.0) or (self.meta['error_ra'] == 0.0):
            warnings.warn('error_ra and/or error_dec is 0.0, please update the value manually in '
                          '"tableOccult_update.txt" before submitting the file')
        f.write(ow_occ_head.format(name=self.meta['name'], ephem=self.meta.get('ephem', 'ephem'),
                                   max_ca=self.meta['max_ca'].to(u.arcsec), size=len(self),
                                   radius=self.meta['radius'], ow_des=ow_des))
        for time, coord, coord_geo, ca, pa, vel, dist, mag, mag_20, longi, loct, md, sd, source in self.iterrows():
            dmag = mag_20-mag
            f.write('{} {} {}  {}   {}   {}   {:5.3f}  {:6.2f} {:-7.3f} {:7.3f} '
                    '{:4.1f} {:-4.1f}   {:3.0f}. {}  {:4.0f}  {:4.0f}\n'.
                    format(time.iso[8:10], time.iso[5:7], time.iso[:4], time.iso[11:20].replace(':', ' '),
                           coord.to_string('hmsdms', precision=4, sep=' ')[:-1],
                           coord_geo.to_string('hmsdms', precision=4, sep=' ')[:-1],
                           ca, pa, vel, dist, mag_20, dmag, longi, loct,
                           self.meta.get('error_ra', 0), self.meta.get('error_dec', 0))
                    )
        f.write(' '+'-'*148+'\n')
        f.close()

        f = open('LOG.dat', modes[mode])
        t = Time.now()
        for time in self['Epoch']:
            t0 = Time(time.isot.split('T')[0] + ' 00:00:00.0')
            dt = (time-t0).jd*24
            f.write('{} {:5s} {}-{:06.3f}\n'.format(t.isot[:-7], ow_des, time.isot.split('T')[0], dt))
        f.close()

    def plot_occ_map(self, **kwargs):
        basename = kwargs.get('nameimg', None)
        for i in range(len(self)):
            if basename and len(self) > 1:
                kwargs['nameimg'] = f'{basename}_{i}'
            self[i].plot_occ_map(**kwargs)

    plot_occ_map.__doc__ = PredictRow.plot_occ_map.__doc__

    def remove_occ(self, date):
        """Removes stellar occultations from table.

        Parameters
        ----------
        date : `str`, `list`
            Date or list of dates of the occultation to be removed.
            The dates mut be as shown in the 'Epoch' column. If the date is not
            complete, the function will select all occultations that matches the
            given string. For instance, ``date='2020-06'`` will remove all
            occultations from the month of June 2020.
        """
        if type(date) == str:
            date = [date]
        itens = list(set([y for d in date for y in self.__itens_by_epoch(d)]))
        self.remove_rows(itens)

    def keep_from_selected_images(self, path='.'):
        """Keeps predictions which images were not deleted in given path.
        This function uses the name of the images to identify predictions.
        The name must be the automatic one generated by plot_occ_map().
        The format of the image is not relevant.

        Parameters
        ----------
        path : `str`
            Path where images are located.
        """
        itens = []
        for i, tca in enumerate(self['Epoch']):
            name = '{}_{}'.format(self.meta['name'], tca.isot)
            if not glob.glob(os.path.join(path, name)+'*'):
                itens.append(i)
        self.remove_rows(itens)



import warnings

__all__ = ['getBSPfromJPL', 'ephem_kernel', 'ephem_horizons']


def getBSPfromJPL(identifier, initial_date, final_date, email, directory='./'):
    """Downloads BSP files from JPL database.

    BSP files, which have information to generate the ephemeris of the objects,
    will be downloaded and named as (without spaces): '[identifier].bsp'.

    Important
    ---------
    It is not possible to download BSP files for Planets or Satellites.

    Parameters
    ----------
    identifier : `str`, `list`
        Identifier of the object(s).
        It can be the `name`, `number` or `SPK ID`.
        It can also be a list of objects.

        Examples: ``'2137295'``, ``'1999 RB216'``, ``'137295'``,
        ``['2137295', '136199', '1999 RC216', 'Chariklo']``.

    initial_date : `str`
        Date the bsp file is to begin, within span `1900-2100`.

        Examples: ``'2003-02-01'``, ``'2003-3-5'``.

    final_date : `str`
        Date the bsp file is to end, within span [1900-2100].
        Must be more than 32 days later than `initial_date`.

        Examples: ``'2006-01-12'``, ``'2006-1-12'``.

    email : `str`
        User's e-mail contact address.
        Required by JPL web service.

        Example: ``username@user.domain.name``.

    directory : `str`
        Directory path to save the bsp files.
    """
    import pathlib
    import shutil
    import requests
    from datetime import datetime

    date1 = datetime.strptime(initial_date, '%Y-%m-%d')
    date2 = datetime.strptime(final_date, '%Y-%m-%d')
    diff = date2 - date1

    if diff.days <= 32:
        raise ValueError('The [final_date] must be more than 32 days later than [initial_date]')
    else:
        if isinstance(identifier, str):
            identifier = [identifier]

        url_jpl = 'https://ssd.jpl.nasa.gov/x/smb_spk.cgi?OPTION=Make+SPK'
        lim, opt, n = 10, 'y', len(identifier)

        if n > lim:
            parameters = {'OBJECT': identifier[0], 'START': date1.strftime('%Y-%b-%d'),
                          'STOP': date2.strftime('%Y-%b-%d'), 'EMAIL': email, 'TYPE': '-B'}

            t0 = datetime.now()
            r = requests.get(url_jpl, params=parameters, stream=True)
            tf = datetime.now()

            bsp_format = r.headers['Content-Type']
            if r.status_code == requests.codes.ok and bsp_format == 'application/download':
                size0 = int(r.headers["Content-Length"]) / 1024 / 1024  # MB

                print('Estimated time to download {} ({:.3f} MB) files: {}'.
                      format(n, n * size0, n * (tf - t0)))

                opt = input('\nAre you sure? (y/n):')
            else:
                raise ValueError('It was not able to download the bsp file for object {}\n'.
                                 format(identifier[0]))

        if opt in ['y', 'Y', 'YES', 'yes']:
            if directory != './':
                path = pathlib.Path(directory)
                if not path.exists():
                    raise ValueError('The directory {} does not exist!'.format(path))
                directory += '/'

            print("\nDownloading bsp file(s) ...\n")

            m, size = 0, 0.0
            failed = []

            t0 = datetime.now()
            for obj in identifier:
                filename = obj.replace(' ', '') + '.bsp'

                parameters = {'OBJECT': obj, 'START': date1.strftime('%Y-%b-%d'),
                              'STOP': date2.strftime('%Y-%b-%d'), 'EMAIL': email, 'TYPE': '-B'}

                r = requests.get(url_jpl, params=parameters, stream=True)
                bsp_format = r.headers['Content-Type']
                if r.status_code == requests.codes.ok and bsp_format == 'application/download':
                    size += int(r.headers["Content-Length"]) / 1024 / 1024
                    m += 1
                    with open(directory + filename, 'wb') as f:
                        r.raw.decode_content = True
                        shutil.copyfileobj(r.raw, f)
                else:
                    failed.append(obj)
            tf = datetime.now()

            print("{} ({:.3f} MB) file(s) was/were downloaded".format(m, size))
            print("Download time: {}".format(tf - t0))

            if len(failed) > 0:
                raise ValueError(
                    'It was not able to download the bsp files for next objects: {}'.format(sorted(failed)))


def ephem_kernel(time, target, observer, kernels, output='ephemeris'):
    """Calculates the ephemeris from kernel files.

    Parameters
    ----------
    time : `str`, `astropy.time.Time`
        Reference instant to calculate ephemeris. It can be a string
        in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

    target : `str`
        IAU (kernel) code of the target.

    observer : `str`
        IAU (kernel) code of the observer.

    kernels : `list`, `str`
        List of paths for all the kernels.

    output : `str`
        The output of data. ``ephemeris`` will output the observed position,
        while ``vector`` will output the Cartesian state vector, without
        light time correction.

    Returns
    -------
    coord : `astropy.coordinates.SkyCoord`
        ICRS coordinate of the target.
    """
    import numpy as np
    import astropy.units as u
    import astropy.constants as const
    import spiceypy as spice

    from astropy.coordinates import SkyCoord
    from astropy.time import Time
    from sora.observer import Observer, Spacecraft

    origins = {'geocenter': '399', 'barycenter': '0'}
    location = origins.get(observer)
    if not location and isinstance(observer, str):
        location = observer
    if isinstance(observer, (Observer, Spacecraft)):
        location = str(getattr(observer, "spkid", None))
    if not location:
        raise ValueError("observer must be 'geocenter', 'barycenter' or an observer object.")
    if output not in ['ephemeris', 'vector']:
        raise ValueError("output must be 'ephemeris' or 'vector'")

    if type(kernels) == str:
        kernels = [kernels]
    for kern in kernels:
        spice.furnsh(kern)
    time = Time(time)
    t0 = Time('J2000', scale='tdb')
    dt = (time - t0)
    delt = 0 * u.s

    # calculates vector Solar System Barycenter -> Observer
    if isinstance(observer, (Observer, Spacecraft)):
        spice.kclear()  # necessary because observer.get_vector() may load different kernels
        position1 = observer.get_vector(time=time, origin='barycenter')
        for kern in kernels:
            spice.furnsh(kern)
    else:
        position1 = spice.spkpos(location, dt.sec, 'J2000', 'NONE', '0')[0]
        position1 = SkyCoord(*position1.T * u.km, representation_type='cartesian')

    while True:
        # calculates new time
        tempo = dt - delt
        # calculates vector Solar System Barycenter -> Object
        position2 = spice.spkpos(target, tempo.sec, 'J2000', 'NONE', '0')[0]
        position2 = SkyCoord(*position2.T * u.km, representation_type='cartesian')
        position = position2.cartesian - position1.cartesian
        # calculates new light time
        delt = (position.norm() / const.c).decompose()
        # if difference between new and previous light time is smaller than 0.001 sec, then continue.
        if output == 'vector' or np.all(np.absolute(((dt - tempo) - delt).sec) < 0.001):
            break
    coord = SkyCoord(position, representation_type='cartesian')
    spice.kclear()
    if output == 'ephemeris':
        coord = SkyCoord(ra=coord.spherical.lon, dec=coord.spherical.lat,
                         distance=coord.spherical.distance, obstime=time)
    if not coord.isscalar and len(coord) == 1:
        coord = coord[0]
    return coord


def ephem_horizons(time, target, observer, id_type='smallbody', output='ephemeris'):
    """Calculates the ephemeris from Horizons.

    Parameters
    ----------
    time : `str`, `astropy.time.Time`
        Reference instant to calculate ephemeris. It can be a string
        in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

    target : `str`
        IAU (kernel) code of the target.

    observer : `str`
        IAU (kernel) code of the observer.

    id_type : `str`
        Type of target object options: ``smallbody``, ``majorbody`` (planets but
        also anything that is not a small body), ``designation``, ``name``,
        ``asteroid_name``, ``comet_name``, ``id`` (Horizons id number), or
        ``smallbody`` (find the closest match under any id_type).

    output : `str`
        The output of data. ``ephemeris`` will output the observed position,
        while ``vector`` will output the Cartesian state vector, without
        light time correction.

    Returns
    -------
    coord : `astropy.coordinates.SkyCoord`
        ICRS coordinate of the target.

    Notes
    -----
    If the interval of time is larger than 30 days or so, a timeout error may be raised.
    The maximum interval will depend on the user connection.
    """
    import astropy.units as u

    from astroquery.jplhorizons import Horizons
    from astropy.time import Time
    from astropy.coordinates import SkyCoord
    from sora.observer import Observer, Spacecraft
    from scipy import interpolate

    origins = {'geocenter': '@399', 'barycenter': '@0'}
    location = origins.get(observer)
    if not location and isinstance(observer, str):
        location = observer
    if isinstance(observer, (Observer, Spacecraft)):
        location = f'{getattr(observer, "code", "")}@{getattr(observer, "spkid", "")}'
    if not location:
        raise ValueError("observer must be 'geocenter', 'barycenter' or an observer object.")
    if output not in ['ephemeris', 'vector']:
        raise ValueError("output must be 'ephemeris' or 'vector'")

    time = Time(time)
    time1 = getattr(time, {'ephemeris': 'utc', 'vector': 'tdb'}[output]).jd
    if not time.isscalar and len(time) > 50:
        step = '10m'
        if time.max() - time.min() > 30 * u.day:
            warnings.warn('Time interval may be too long. A timeout error may be raised.')
        if time.max() - time.min() <= 1 * u.day:
            step = '1m'
        time2 = {'start': (time.min() - 10 * u.min).iso.split('.')[0],
                 'stop': (time.max() + 10 * u.min).iso.split('.')[0],
                 'step': step,
                 }
    else:
        time2 = time1

    if getattr(observer, 'ephem', None) not in ['horizons', None]:
        warnings.warn('Ephemeris using kernel for the observer and Horizons for the target is under construction. '
                      'We will use only Horizons.')
    ob = Horizons(id=target, id_type=id_type, location=location, epochs=time2)

    if output == 'ephemeris':
        eph = ob.ephemerides(extra_precision=True)
        obstime = Time(eph['datetime_jd'], format='jd', scale='utc')
        pos = SkyCoord(eph['RA'], eph['DEC'], eph['delta'], frame='icrs', obstime=obstime)
    else:
        vec = ob.vectors(refplane='earth')
        obstime = Time(vec['datetime_jd'], format='jd', scale='tdb')
        pos = SkyCoord(*[vec[i] for i in ['x', 'y', 'z']] * u.AU, representation_type='cartesian', obstime=obstime)

    if isinstance(time2, dict):
        spl_x = interpolate.CubicSpline(obstime.jd, pos.cartesian.x.to(u.km))
        spl_y = interpolate.CubicSpline(obstime.jd, pos.cartesian.y.to(u.km))
        spl_z = interpolate.CubicSpline(obstime.jd, pos.cartesian.z.to(u.km))
        pos = SkyCoord(x=spl_x(time1), y=spl_y(time1), z=spl_z(time1), unit=u.km, representation_type='cartesian')

    if output == 'ephemeris':
        pos = SkyCoord(ra=pos.spherical.lon, dec=pos.spherical.lat, distance=pos.spherical.distance)

    if not pos.isscalar and len(pos) == 1:
        pos = pos[0]

    return pos



not_found_message = 'specified object was not found'
many_objects_message = 'specified query matched more than one object'

spectral_features = {
    "id1": "Very steep red slope shortward of 0.75 μm; moderately deep absorption feature longward of 0.75 μm.",
    "id2": ("Linear, generally featureless spectra. Differences in UV absorption features and presence/absence"
            " of narrow absorption feature near 0.7 μm."),
    "id3": ("Linear, generally featureless spectra. Differences in UV absorption features and presence/absence"
            " of narrow absorption feature near 0.7 μm."),
    "id4": "Relatively featureless spectrum with very steep red slope.",
    "id5": ("Generally featureless spectrum with reddish slope; differences in subtle absorption features"
            " and/or spectral curvature and/or peak relative reflectance."),
    "id6": "Reddish slope shortward of 0.7 μm; deep, rounded absorption feature longward of 0.75 μm.",
    "id7": "Moderate reddish slope downward of 0.7 μm; deep absorption longward of 0.75 μm.",
    "id8": ("Moderately steep reddish slope downward of 0.7 μm; moderate to steep absorption longward of 0.75 μm;"
            " peak of reflectance at 0.73 μm. Bus subgroups intermediate between S and A, K, L, Q, R classes."),
    "id9": "Moderately reddish shortward of 0.75 μm; flat afterward.",
    "id10": "Reddish shortward of 0.7 μm; extremely deep absorption longward of 0.75 μm.",
    "id11": ("Moderately steep red slope shortward of 0.75 μm; smoothly angled maximum and flat to blueish"
             " longward of 0.75 μm, with little or no curvature."),
    "id12": "Very steep red slope shortward of 0.75 μm; flat longward of 0.75 μm; differences in peak level.",
    "id13": "Peculiar trend, known so far for very few asteroids."
}

smass = {
    "A": spectral_features["id1"],
    "B": spectral_features["id2"],
    "C": spectral_features["id3"],
    "Cb": spectral_features["id3"],
    "Ch": spectral_features["id3"],
    "Cg": spectral_features["id3"],
    "Chg": spectral_features["id3"],
    "D": spectral_features["id4"],
    "X": spectral_features["id5"],
    "Xc": spectral_features["id5"],
    "Xe": spectral_features["id5"],
    "Xk": spectral_features["id5"],
    "Q": spectral_features["id6"],
    "R": spectral_features["id7"],
    "S": spectral_features["id8"],
    "Sa": spectral_features["id8"],
    "Sk": spectral_features["id8"],
    "Sl": spectral_features["id8"],
    "Sq": spectral_features["id8"],
    "Sr": spectral_features["id8"],
    "T": spectral_features["id9"],
    "V": spectral_features["id10"],
    "K": spectral_features["id11"],
    "L": spectral_features["id12"],
    "Ld": spectral_features["id12"],
    "O": spectral_features["id13"]
}

tholen = {
    "A": spectral_features["id1"],
    "B": spectral_features["id2"],
    "F": spectral_features["id2"],
    "C": spectral_features["id3"],
    "G": spectral_features["id3"],
    "D": spectral_features["id4"],
    "E": spectral_features["id5"],
    "M": spectral_features["id5"],
    "P": spectral_features["id5"],
    "Q": spectral_features["id6"],
    "R": spectral_features["id7"],
    "S": spectral_features["id8"],
    "T": spectral_features["id9"],
    "V": spectral_features["id10"]
}



import warnings

import numpy as np

from sora.config.list import List
from .chord import Chord

__all__ = ['ChordList']


class ChordList(List):
    """Defines a collection of Chord objects associated to an Occultation.

    This object is not supposed to be defined by the user. It will be automatically
    defined in Occultation.

    Attributes
    ----------
    star : `sora.Star`
        The Star occulted.

    body : `sora.Body`
        The occulting Body.

    time : `astropy.time.Time`
        The occultation time.

    """
    _allowed_types = (Chord,)  # (Chord, AstrometricChord)
    _set_func = 'add_chord'

    def __init__(self, *, star, body, time):

        super().__init__()
        self._star = star
        self._body = body
        self._time = time
        self._shared_with = {"chord": {"star": self._star, "ephem": self._body.ephem, "time": self._time},
                             'occultation': {}}
        self._method_value = 'geocenter'

    def add_chord(self, *, name=None, chord=None, observer=None, lightcurve=None):
        """Add a chord to the occultation chord list

        Parameters
        ----------
        name : `str`
            The name of the Chord. It must be unique within the list of chords.
            If not given, it will use the name in chord if chord is directly
            given or the name in the observer object. If the name in Chord already
            exists in the list, the name parameter can be given to update the
            chord name. The `name` parameter is required if given together with
            observer and lightcurve. It can not be an empty string.

        chord `sora.occultation.Chord`
            A chord instance defined by the user.

        observer : `sora.Observer`
            An Observer instance defined by the user. It must be given together
            with a lightcurve.

        lightcurve : `sora.LightCurve`
            A lightcurve defined by the user. It must be given together with observer.

        Examples
        --------
        Ways to add a chord:

        >>> obj.add_chord(chord) # where the name in chord will be used.

        >>> obj.add_chord(name, chord) # where the name in chord will be replaced by "name".

        >>> obj.add_chord(observer, lightcurve) # where the name in observer will be used.

        >>> obj.add_chord(name, observer, lightcurve)
        """
        if chord and (observer or lightcurve):
            raise ValueError("User must give only chord or (name and observer and lightcurve)")
        if chord:
            name = name or chord.name
        elif observer and lightcurve:
            for key in self.keys():
                if self[key].lightcurve is lightcurve:
                    raise ValueError('lightcurve is already associated to the chord {}'.format(key))
            name = name or observer.name
            chord = Chord(name=name, observer=observer, lightcurve=lightcurve)
        else:
            raise ValueError("User must give chord or (name and observer and lightcurve)")
        self._add_item(name=name, item=chord)
        chord._name = name
        chord._shared_with["chordlist"] = self._shared_with["chord"]
        chord.lightcurve.set_vel(self._shared_with['occultation']["vel"])
        chord.lightcurve.set_dist(self._shared_with['occultation']["dist"])
        chord.lightcurve.set_star_diam(self._shared_with['occultation']["star_diam"])
        try:
            chord.lightcurve.calc_magnitude_drop(mag_star=self._star.mag['G'], mag_obj=self._body.apparent_magnitude(self._time))
        except:
            chord.lightcurve.bottom_flux = 0.0
            warnings.warn('Magnitude drop was not calculated. Using bottom flux as 0.0.')
        chord._method = self._method
        return chord

    # This attribute is to modify the way to calculate f and g on chord.
    @property
    def _method(self):
        return self._method_value

    @_method.setter
    def _method(self, value):
        if value not in ['geocenter', 'observer']:
            raise ValueError('method must be "geocenter" or "observer"')
        self._method_value = value
        for name, chord in self.items():
            chord._method = value

    def remove_chord(self, *, name):
        """Remove a chord from the chord list and disassociate it from the Occultation.

        Parameters
        ----------
        name : `str`
            The name of the chord.
        """
        del(self[name]._shared_with['chordlist'])
        del(self[name])

    @property
    def is_able(self):
        return {name: chord.is_able for name, chord in self.items()}

    def enable(self, *, chord=None, time=None):
        """Enable a contact point of the curve to be used in the fit.

        Parameters
        ----------
        chord : `str`
            Name of the chord to enable. If ``chord=None``, it applies to all chords.

        time : Non`, `str`
            If ``time=None``, it will enable all contact points.
            If 'immersion' or 'emersion', it will enable respective contact point.
        """
        n = 0
        for key in self.keys():
            if chord is None or chord == key:
                self[key].enable(time=time)
                n += 1
        if n == 0:
            raise ValueError('No Chord with name {} is available.'.format(chord))

    def disable(self, *, chord=None, time=None):
        """Disable a contact point of the curve to be used in the fit.

        Parameters
        ----------
        chord : `str`
            Name of the chord to disable. If ``chord=None``, it applies to all chords.

        time : None, `str`
            If ``time=None``, it will disable all contact points.
            If 'immersion' or 'emersion', it will disable respective contact point.
        """
        n = 0
        for key in self.keys():
            if chord is None or chord == key:
                self[key].disable(time=time)
                n += 1
        if n == 0:
            raise ValueError('No Chord with name {} is available.'.format(chord))

    def plot_chords(self, *, segment='standard', ignore_chords=None, only_able=False, ax=None, linestyle='-', **kwargs):
        """Plots the on-sky path of this chord.

        Parameters
        ----------
        segment : `str`
            The segment to plot the chord. The available options are:

            ``'positive'`` to get the path between the immersion and emersion
            times if the chord is positive.

            ``'negative'`` to get the path between the start and end of
            observation if the chord is negative.

            ``'standard'`` to get the 'positive' path if the chord is positive
            or 'negative' if the chord is negative.

            ``'full'`` to get  the path between the start and end of observation
            independent if the chord is positive or negative.

            ``'outer'`` to get the path outside the 'positive' path, for instance
            between the start and immersion times and between the emersion and
            end times.

            ``'error'`` to get the path corresponding to the error bars.

        ignore_chords : `str`, `list`
            Name of chord or list of names to ignore in the plot.

        only_able : `bool`
            Plot only the chords or contact points that are able to be used in the fit.
            If ``segment='error'`` it will show only the contact points able.
            If segment is any other, the path will be plotted only if both
            immersion and emersion are able, or it is a negative chord.

        ax : `matplotlib.pyplot.Axes`
            The axes where to make the plot. If None, it will use the default axes.

        linestyle : `str`
            Default linestyle used in `matplotlib.pyplot.plot`. The difference
            is that now it accept ``linestyle='exposure'``, where the plot will
            be a dashed line corresponding to each exposure. The blank space
            between the lines can be interpreted as 'dead time'.

        **kwargs
            Any other kwarg will be parsed directly by `maplotlip.pyplot.plot`.
            The only difference is that the default linewidth ``lw=2``.

        """
        n = 0
        keys = list(self.keys())
        if ignore_chords is not None:
            ignore_chords = np.array(ignore_chords, ndmin=1)
        for i in range(len(self)):
            if ignore_chords is not None and keys[i] in ignore_chords:
                continue
            if segment != 'error':
                kwargs['label'] = keys[i]
            try:
                _ = self[i].plot_chord(segment=segment, only_able=only_able, ax=ax, linestyle=linestyle, **kwargs)
            except ValueError:
                n += 1
        if n == len(self):
            warnings.warn('Segment "{}" was not found on any chord'.format(segment))

    def summary(self):
        """Prints a table with the summary of the chords.
        """
        from astropy.table import Table, vstack

        tables = []
        for key in self.keys():
            tt = Table()
            obs = self[key].observer
            lc = self[key].lightcurve
            max_val = 0
            cols = []
            colnames = ['Name', 'Longitude', 'Latitude', 'status', 'time', 'f', 'g']
            itens = [key, obs.lon.to_string(), obs.lat.to_string()]
            row = []
            times = {'Initial Time': 'initial_time', 'Immersion': 'immersion', 'Emersion': 'emersion', 'End Time': 'end_time'}
            for i in times:
                val = getattr(lc, times[i], None)
                if val is not None:
                    row.append([i, val.iso, *['{:.2f}'.format(n) for n in self[key].get_fg(time=val)]])
            for t in np.array(row).T:
                itens.append(t.tolist())
            for item in itens:
                v = np.array(item, ndmin=1).tolist()
                cols.append(v)
                if len(v) > max_val:
                    max_val = len(v)
            for i in range(len(cols)):
                if len(cols[i]) < max_val:
                    for n in range(max_val - len(cols[i])):
                        cols[i].append('')
                tt[colnames[i]] = cols[i]
            tables.append(tt)
        tabela = vstack(tables)
        tabela.pprint_all()

    def get_impact_param(self, chords='all_chords', center_f=0, center_g=0, verbose=True):
        """Get the impact parameter, minimal distance between the chord and the
        centre position.

        This Chord object must be associated to an Occultation to work, since it
        needs the position of the star and an ephemeris.

        Parameters
        ----------
        chords : `int`, `str`, default='all_chords'
            Index or names of the chords to be considered.

        center_f : `int`, `float`, default=0
            The coordinate in f of the ellipse center.

        center_g : `int`, `float`, default=0
            The coordinate in g of the ellipse center.

        verbose : `bool`
            If True, prints the obtained values.


        Returns
        -------
        impact, sense, chord_name : `list`
            The impact parameter (in km), the direction of the chord relative
            the ellipse center, North (N), South (S), East (E) and West (W), and
            the name of the chord
        """
        impact = np.array([])
        sense = np.array([])
        names = np.array([])
        if chords == 'all_chords':
            chords = range(len(self))
        for i in chords:
            chord = self[i]
            im, se = chord.get_impact_param(center_f=center_f, center_g=center_g, verbose=verbose)
            impact = np.append(impact, im)
            sense = np.append(sense, se)
            names = np.append(names, chord.name)
        return impact, sense, names

    def get_theoretical_times(self, equatorial_radius, chords='all_chords', center_f=0, center_g=0, oblateness=0,
                              position_angle=0, sigma=0, step=1, verbose=True):
        """Get the theoretical times and chords sizes for a given ellipse.

        This Chord object must be associated to an Occultation to work, since it needs
        the position of the star and an ephemeris.

        Parameters
        ----------
        chords : `int`, `str`, default='all_chords'
            Index or names of the chords to be considered.

        equatorial_radius : `int`, `float`
            The Equatorial radius (semi-major axis) of the ellipse.

        center_f : `int`, `float`, default=0
            The coordinate in f of the ellipse center.

        center_g : `int`, `float`, default=0
            The coordinate in g of the ellipse center.

        oblateness : `int`, `float`, default=0
            The oblateness of the ellipse.

        position_angle : `int`, `float`, default=0
            The pole position angle of the ellipse in degrees.
            Zero is in the North direction ('g-positive'). Positive clockwise.

        sigma : `int`, `float`
            Uncertainty of the expected ellipse, in km.

        step : `int`, `float`
            Time resolution of the chord, in seconds.

        verbose : `bool`
            If True, prints the obtained values.


        Returns
        -------
        theory_immersion_time, theory_emersion_time, theory_chord_size, chord_name : `list`
            The expected immersion time for the given ellipse, the expected
            emersion time for the given ellipse, the expected chord size for the
            given ellipse, and the name of the chord.
        """
        theory_immersion_time = np.array([])
        theory_emersion_time = np.array([])
        theory_chord_size = np.array([])
        names = np.array([])
        if chords == 'all_chords':
            chords = range(len(self))
        for i in chords:
            chord = self[i]
            tit, tet, tcs = chord.get_theoretical_times(equatorial_radius=equatorial_radius, center_f=center_f,
                                                        center_g=center_g, oblateness=oblateness, position_angle=position_angle,
                                                        sigma=sigma, step=step, verbose=verbose)
            theory_immersion_time = np.append(theory_immersion_time, tit)
            theory_emersion_time = np.append(theory_emersion_time, tet)
            theory_chord_size = np.append(theory_chord_size, tcs)
            names = np.append(names, chord.name)
        return theory_immersion_time, theory_emersion_time, theory_chord_size, names



import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord, Longitude, Latitude

from sora.ephem import EphemPlanete, EphemKernel, EphemJPL, EphemHorizons

__all__ = ['PhysicalData']


class PhysicalData(u.Quantity):
    """Defines PhysicalData with uncertainty, reference and notes.

    Note
    ----
    It inherits from astropy.units.quantity.Quantity().

    Parameters
    ----------
    name :`str`
        The name representing the corresponding physical parameter.

    value : `int`, `float`, `str`, `~numpy.ndarray`, `astropy.quantity.Quantity`
        The numerical value of this quantity in the units given by unit.  If
        a `Quantity`  (or any other valid object with a ``unit`` attribute),
        creates a new `Quantity` object, converting to `unit` units as needed.
        If a string, it is converted to a number or `Quantity`, depending on
        whether a unit is present.

    uncertainty : `int`, `float`, `str`, `~numpy.ndarray`, `astropy.quantity.Quantity`, default=0
        The numerical value of this quantity in the units given by unit.  If
        a `Quantity` (or any other valid object with a ``unit`` attribute),
        creates a new `Quantity` object, converting to `unit` units as needed.
        If a string, it is converted to a number or `Quantity`, depending on
        whether a unit is present.

    reference : `str`, default='User'
        A string stating the reference for the parameter value.

    notes : `str`, default=''
        Any other important information about the physical parameter.

    unit : `str`, `~astropy.units.UnitBase` instance, default='dimensionless'
        An object that represents the unit associated with the input value. Must
        be an `~astropy.units.UnitBase` object or a string parsable by the
        :mod:`~astropy.units` package.

    raise_error : `bool`, default=False
        If ``value=None`` or ``raise_error=True`` the function raises an error,
        else `value` is redefined to ``NaN``.

    """

    def __new__(cls, name, value, uncertainty=0.0, reference="User", notes="", unit=u.dimensionless_unscaled,
                raise_error=False):
        given_unit = unit
        if value is None:
            if raise_error:
                raise TypeError("The value must be a valid PhysicalData type. Given: {}".format(value))
            else:
                value = np.nan
        elif isinstance(value, u.core.CompositeUnit):
            unit = u.dimensionless_unscaled
            for i in np.arange(len(value.bases)):
                unit = unit*(value.bases[i]**value.powers[i])
            value = value.scale
        elif isinstance(value, (PhysicalData, u.Quantity)):
            unit = value.unit
            value = value.value
        if not np.isscalar(value):
            print(value)
            raise ValueError('Given value must be a scalar. Given: {}'.format(value))
        if not unit.is_equivalent(given_unit):
            raise ValueError('{} is not equivalent to {}'.format(unit, given_unit))
        physdata = super().__new__(cls, value=value, unit=unit)
        physdata = physdata.to(given_unit)
        physdata.name = name
        physdata.uncertainty = uncertainty
        physdata.reference = reference
        physdata.notes = notes
        return physdata

    @property
    def uncertainty(self):
        return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, value):
        unit = self.unit
        given_unit = unit
        if isinstance(value, u.core.CompositeUnit):
            unit = u.dimensionless_unscaled
            for i in np.arange(len(value.bases)):
                unit = unit*(value.bases[i]**value.powers[i])
            value = value.scale
        elif value is None:
            value = 0.0
        elif isinstance(value, u.Quantity):
            unit = value.unit
            value = value.value
        if not np.isscalar(value):
            print(value)
            raise ValueError('Given value must be a scalar. Given: {}'.format(value))
        if not unit.is_equivalent(self.unit):
            raise ValueError('{} is not equivalent to {}'.format(unit, given_unit))
        self._uncertainty = u.Quantity(value, unit).to(given_unit)
        if self._uncertainty < 0:
            raise ValueError('uncertainty cannot be a negative value')

    @property
    def reference(self):
        return self._reference

    @reference.setter
    def reference(self, value):
        if value is None:
            value = ""
        if not isinstance(value, str):
            raise TypeError('reference must be a string')
        self._reference = value

    @property
    def notes(self):
        return self._notes

    @notes.setter
    def notes(self, value):
        if value is None:
            value = ""
        if not isinstance(value, str):
            raise TypeError('notes must be a string')
        self._notes = value

    def __repr__(self):
        if not np.isnan(self.value):
            return super().__repr__()
        else:
            return ""

    def __str__(self):
        if not np.isnan(self.value):
            out = "{}:\n    {:.5g} +/- {:.5g} {}\n    Reference: {}, {}\n".format(
                self.name, self.value, self.uncertainty.value, self.unit.__str__(), self.reference, self.notes)
            return out
        else:
            return ""


class BaseBody():

    @property
    def albedo(self):
        return self._albedo

    @albedo.setter
    def albedo(self, value):
        if isinstance(value, PhysicalData):
            albedo = value
            albedo.name = "Albedo"
        else:
            albedo = PhysicalData('Albedo', value)
        if albedo < 0:
            raise ValueError("albedo cannot be a negative value")
        self._albedo = albedo

    @property
    def H(self):
        return self._H

    @H.setter
    def H(self, value):
        if isinstance(value, PhysicalData):
            self._H = value
        else:
            self._H = PhysicalData('Absolute Magnitude', value, unit=u.mag)
        if not np.isnan(self.H):  # remove this line for v1.0
            self._shared_with['ephem']['H'] = self.H  # remove this line for v1.0

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, value):
        if isinstance(value, PhysicalData):
            self._G = value
        else:
            self._G = PhysicalData('Phase Slope', value)
        if not np.isnan(self._G):  # remove this line for v1.0
            self._shared_with['ephem']['G'] = self.G  # remove this line for v1.0

    @property
    def diameter(self):
        return self._diameter

    @diameter.setter
    def diameter(self, value):
        if isinstance(value, PhysicalData):
            diameter = value
            diameter.name = "Diameter"
        else:
            diameter = PhysicalData('Diameter', value, unit=u.km)
        if diameter < 0:
            raise ValueError("diameter cannot be a negative value")
        self._diameter = diameter
        self._shared_with['ephem']['radius'] = self.radius

    @property
    def radius(self):
        return PhysicalData('Radius', self.diameter/2.0, self.diameter.uncertainty/2,
                            self.diameter.reference, self.diameter.notes, unit=u.km)

    @radius.setter
    def radius(self, value):
        if isinstance(value, PhysicalData):
            self.diameter = value*2.0
            self.diameter.uncertainty = value.uncertainty * 2.0
            self.diameter.name = 'Diameter'
        else:
            self.diameter = PhysicalData('Diameter', float(value) * 2.0, unit=u.km)

    @property
    def density(self):
        return self._density

    @density.setter
    def density(self, value):
        if isinstance(value, PhysicalData):
            density = value
            density.name = "Density"
        else:
            density = PhysicalData('Density', value, unit=u.g / u.cm ** 3)
        if density < 0:
            raise ValueError("density cannot be a negative value")
        self._density = density

    @property
    def GM(self):
        return self._GM

    @GM.setter
    def GM(self, value):
        if isinstance(value, PhysicalData):
            GM = value
            GM.name = 'Standard Gravitational Parameter'
        else:
            GM = PhysicalData('Standard Gravitational Parameter', value, unit=u.km ** 3 / u.s ** 2)
        if GM < 0:
            raise ValueError("GM cannot be a negative value")
        self._GM = GM

    @property
    def mass(self):
        import astropy.constants as const

        return PhysicalData('Mass', self.GM/const.G, self.GM.uncertainty/const.G,
                            self.GM.reference, self.GM.notes, unit=u.kg)

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        if isinstance(value, PhysicalData):
            rotation = value
            rotation.name = "Rotation"
        else:
            rotation = PhysicalData('Rotation', value, unit=u.h)
        if rotation < 0:
            raise ValueError("rotation cannot be a negative value")
        self._rotation = rotation

    @property
    def pole(self):
        return self._pole

    @pole.setter
    def pole(self, value):
        if value is None:
            self._pole = SkyCoord(np.nan, np.nan, unit=(u.deg, u.deg))
        else:
            self._pole = SkyCoord(value, unit=(u.hourangle, u.deg))
        self._pole.ra.uncertainty = Longitude(0*u.hourangle)
        self._pole.dec.uncertainty = Latitude(0*u.deg)
        self._pole.reference = "User"
        self._pole.notes = ""

    @property
    def BV(self):
        return self._BV

    @BV.setter
    def BV(self, value):
        if isinstance(value, PhysicalData):
            self._BV = value
        else:
            self._BV = PhysicalData('B-V color', value)

    @property
    def UB(self):
        return self._UB

    @UB.setter
    def UB(self, value):
        if isinstance(value, PhysicalData):
            self._UB = value
        else:
            self._UB = PhysicalData('U-B color', value)

    @property
    def spkid(self):
        return self._shared_with['ephem'].get('spkid')

    @spkid.setter
    def spkid(self, value):
        spkid = str(int(value))
        self._shared_with['ephem']['spkid'] = spkid

    @property
    def orbit_class(self):
        return self._orbit_class

    @orbit_class.setter
    def orbit_class(self, value):
        orbit_classes = {'tno': 'TransNeptunian Object', 'satellite': 'Natural Satellite', 'centaur': 'Centaur',
                         'comet': 'Comet', 'asteroid': 'Main-belt Asteroid', 'trojan': 'Jupiter Trojan',
                         'neo': 'Near-Earth Object', 'planet': "Planet", 'unclassified': "Unclassified"}
        if value in orbit_classes.keys():
            self._orbit_class = orbit_classes[value.lower()]
        elif value in orbit_classes.values():
            self._orbit_class = value
        else:
            self._orbit_class = orbit_classes['unclassified']

    @property
    def ephem(self):
        try:
            return self._ephem
        except AttributeError:
            raise AttributeError('{} object does not have ephemeris.'.format(self.__class__.__name__))

    @ephem.setter
    def ephem(self, value):
        allowed_types = [EphemPlanete, EphemKernel, EphemJPL, EphemHorizons]
        if type(value) not in allowed_types:
            if isinstance(value, str) and value.lower() == 'horizons':
                value = EphemHorizons(name=self._search_name, id_type=self._id_type, spkid=self.spkid)
            elif isinstance(value, (list, str)):
                value = EphemKernel(kernels=value, spkid=self.spkid)
            else:
                raise ValueError('Cannot set "ephem" with {}. Allowed types are: {}'.format(type(value), allowed_types))
        if 'spkid' not in self._shared_with['ephem'] or self._shared_with['ephem']['spkid'] is None:
            if hasattr(value, '_spkid'):
                self._shared_with['ephem']['spkid'] = value.spkid
            else:
                raise AttributeError('spkid is not defined in {} or {}'.format(self.__class__.__name__, value.__class__.__name__))
        if hasattr(self, '_ephem'):
            self._ephem._shared_with['body'] = {}
        self._ephem = value
        spkval = getattr(value, 'spkid', None)
        self._ephem._shared_with['body'] = self._shared_with['ephem']
        spknewval = getattr(value, 'spkid', None)
        if spkval != spknewval:
            warnings.warn('spkid is different in {0} ({1}) and {2} ({3}). {0}\'s spkid will have higher priority'.format(
                self.__class__.__name__, spknewval, value.__class__.__name__, spkval))

    @property
    def _search_name(self):
        if self.orbit_class in ['Natural Satellite', 'Planet']:
            return str(self.spkid or self.shortname)
        elif hasattr(self, '_des_name'):
            return self._des_name
        else:
            return self.name

    @property
    def _id_type(self):
        if self.orbit_class in ['Natural Satellite', 'Planet']:
            return 'majorbody'
        else:
            return 'smallbody'



import astropy.units as u
import numpy as np

__all__ = ['search_sbdb', 'apparent_magnitude']


def search_sbdb(name):
    """Searches JPL Small-Body DataBase to search for object information.

    As the name implies, it searches only for Small Bodies. Planets and
    satellites information are not retrieved by this function.

    Parameters
    ----------
    name : `str`
        The name of the object for the search. It can be the attributed `spkid`
        or `designation number`. The name is case insensitive.

    Returns
    -------
    sbdb : `dict`
        An ordered dictionary with the object information.

    Important
    ---------
    The query is not an autocomplete search, so ``name='Charikl'`` will not find
    Chariklo. If more than 1 object is found, the user is asked to select the
    correct one (e.g: ``name='neowise'``).
    """
    from astroquery.jplsbdb import SBDB
    from . import values

    print('Obtaining data for {} from SBDB'.format(name))
    sbdb = SBDB.query(name, full_precision=True, solution_epoch=True, validity=True, phys=True, discovery=True)
    if 'message' in sbdb:
        if sbdb['message'] == values.not_found_message:
            raise ValueError(values.not_found_message + " on SBDB")
        elif sbdb['message'] == values.many_objects_message:
            sbdb = select_body(sbdb)
    return sbdb


def select_body(sbdb):
    """Creates an object selection table.

    A table to select object is created when the SBDB search returns more than
    one object. This function is not supposed to be called by the user.

    Parameters
    ----------
    sbdb : `dict`
        An ordered dictionary object returned by SBDB search.

    Returns
    -------
    sbdb : `dict`
        An ordered dictionary with the data of the selected object.
    """
    from astropy.table import Table

    print('{} bodies were found.'.format(sbdb['count']))
    t = Table()
    t['num'] = np.arange(sbdb['count']) + 1
    t['Designation'] = sbdb['list']['pdes']
    t['Name'] = sbdb['list']['name']
    while True:
        t.pprint_all()
        print('0: Cancel')
        choice = int(input('Choose the corresponding number of the correct small body: '))
        if choice in np.arange(sbdb['count'] + 1):
            break
    if choice == 0:
        raise ValueError('It was not possible to define a Small Body')
    return search_sbdb(name=sbdb['list']['pdes'][choice - 1])


def apparent_magnitude(H, G, dist, sundist, phase=0.0):
    """Calculates the object's apparent magnitude.

    Parameters
    ----------
    H : `int`, `float`
        Absolute magnitude.

    G : `int`, `float`
        Slope parameter.

    dist : `int`, `float`
        Observer-object distance, in AU.

    sundist : `int`, `float`
        Sun-object distance, in AU.

    phase : `int`, `float`, default=0
        Phase angle (Sun-Target-Observer), in degrees.


    Returns
    -------
    ap_mag : `float`
        Apparent magnitude.
    """
    phi0 = np.exp(-3.33 * (np.tan(0.5 * phase * u.deg) ** 0.63))
    phi1 = np.exp(-1.87 * (np.tan(0.5 * phase * u.deg) ** 1.22))
    Ha = H - 2.5 * np.log10((1.0 - G) * phi0 + G * phi1)
    ap_mag = Ha + 5 * np.log10(sundist * dist)
    return ap_mag.value


# This block is a hard coded search for satellite information.
# For the moment it only covers some of what SBDB cannot offer.
# It is supposed to be replaced by a search on a trustful
# database on future versions.
grav = 'Grav et al. (2015). AJ 809(1):3'
rettig = 'Rettig et al. (2001). Icarus 154(2):313-320'
bauer = 'Bauer et al. (2006). Icarus 184(1):181-197'
veverka = 'Veverka et al. (1991). book Uranus 528-560'
satellites = {
    'io': {
        'spkid': 501,
        'albedo': [0.63, 0.02, 'Simonelli et al. (1984). Icarus 59:406-425'],
        'diameter': [3643.2, 1.0, 'horizons'],
        'GM': [5959.916, 0.012, 'JUP230']
    },
    'europa': {
        'spkid': 502,
        'albedo': [0.67, 0.03, 'Buratti et al. (1983). Icarus 55:93-110'],
        'diameter': [3121.6, 1.0, 'horizons'],
        'GM': [3202.739, 0.009, 'JUP230']
    },
    'ganymede': {
        'spkid': 503,
        'albedo': [0.43, 0.02, 'Morrison et al. (1977). book Planetary Satellites 363-378'],
        'diameter': [5262.4, 3.4, 'Anderson et al. (2000). BAAS 33(3):1101'],
        'GM': [9887.834, 0.017, 'JUP230']
    },
    'callisto': {
        'spkid': 504,
        'albedo': [0.17, 0.02, 'Morrison et al. (1977). book Planetary Satellites 363-378'],
        'diameter': [4820.6, 3.0, 'Morrison et al. (2000). Icarus 153:157-161'],
        'GM': [7179.289, 0.013, 'JUP230']
    },
    'amalthea': {
        'spkid': 505,
        'albedo': [0.090, 0.005, 'Simonelli et al. (2000). Icarus 147:353-365'],
        'diameter': [166.9, 4.8, 'horizons'],
        'GM': [0.138, 0.03, 'JUP230']
    },
    'thebe': {
        'spkid': 514,
        'albedo': [0.047, 0.003, 'Simonelli et al. (2000). Icarus 147:353-365'],
        'diameter': [98.6, 4.0, 'Thomas et al. (1998). Icarus 135:360-371'],
        'GM': [0.10, 0.0, 'horizons']
    },
    'himalia': {
        'spkid': 506,
        'albedo': [5.7, 0.1, grav],
        'H': [8.0, 0.01, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [139.6, 1.7, grav],
        'GM': [0.28, 0.04, 'Emelyanov (2005). A&A 438(3):L33-L36'],
        'rotation': [7.7819, 0.0005, 'Pilcher et al. (2012). Icarus 219(2):741-742']
    },
    'elara': {
        'spkid': 507,
        'albedo': [5.7, 0.8, grav],
        'H': [9.45, 0.01, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [79.9, 1.7, grav]
    },
    'pasiphae': {
        'spkid': 508,
        'albedo': [4.4, 0.6, grav],
        'H': [10.21, 0.01, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [57.8, 0.8, grav]
    },
    'sinope': {
        'spkid': 509,
        'albedo': [4.2, 0.6, grav],
        'H': [11.29, 0.01, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [35.0, 0.6, grav]
    },
    'lysithea': {
        'spkid': 510,
        'albedo': [3.6, 0.6, grav],
        'H': [11.09, 0.02, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [42.2, 0.7, grav]
    },
    'carme': {
        'spkid': 511,
        'albedo': [3.5, 0.6, grav],
        'H': [10.91, 0.01, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [46.7, 0.9, grav]
    },
    'ananke': {
        'spkid': 512,
        'albedo': [3.8, 0.6, grav],
        'H': [11.84, 0.03, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [29.1, 0.6, grav]
    },
    'leda': {
        'spkid': 513,
        'albedo': [3.4, 0.6, grav],
        'H': [12.63, 0.03, rettig],
        'G': [0.10, 0.15, ''],
        'diameter': [21.5, 1.7, grav]
    },
    'phoebe': {
        'spkid': 609,
        'albedo': [10.0, 0.5, grav],
        'H': [6.59, 0.02, bauer],
        'G': [0.02, 0.03, bauer],
        'diameter': [202.2, 4.5, grav],
        'density': [1.630, 0.045, 'Porco et al. (2005). Science 307(5713):1237-1242'],
        'GM': [0.5534, 0.0006, 'Jacobson et al. (2006). AJ 132(6):2520-2526'],
        'rotation': [9.27365, 0.00002, 'Gomes-Júnior et al. (2020). MNRAS 492(1):770-781'],
        'pole': [['356.6/77.9'], ['0/0'], 'Porco et al. (2005). Science 307(5713):1237-1242']
    },
    'albiorix': {
        'spkid': 626,
        'albedo': [6.2, 2.8, grav],
        'H': [11.35, 0.05, bauer],
        'G': [0.39, 0.06, bauer],
        'diameter': [28.6, 5.4, grav]
    },
    'siarnaq': {
        'spkid': 629,
        'albedo': [5.0, 1.7, grav],
        'H': [10.90, 0.05, bauer],
        'G': [0.45, 0.17, bauer],
        'diameter': [39.3, 5.9, grav]
    },
    'ariel': {
        'spkid': 701,
        'albedo': [0.39, 0.04, veverka],
        'H': [1.3, 0.2, veverka],
        'G': [0.10, 0.15, ''],
        'diameter': [1157.8, 1.2, 'Thomas et al. (1988). Icarus 73:427-441'],
        'GM': [86.4, 5.0, 'Jacobson et al. (2007). BAAS 39:453']
    },
    'umbriel': {
        'spkid': 702,
        'albedo': [0.21, 0.02, veverka],
        'H': [2.2, 0.2, veverka],
        'G': [0.10, 0.15, ''],
        'diameter': [1169.4, 5.6, 'Thomas et al. (1988). Icarus 73:427-441'],
        'GM': [81.5, 5.0, 'Jacobson et al. (2007). BAAS 39:453']
    },
    'titania': {
        'spkid': 703,
        'albedo': [0.27, 0.03, veverka],
        'H': [1.2, 0.2, veverka],
        'G': [0.10, 0.15, ''],
        'diameter': [1577.8, 3.6, 'Thomas et al. (1988). Icarus 73:427-441'],
        'GM': [228.2, 5.0, 'Jacobson et al. (2007). BAAS 39:453']
    },
    'oberon': {
        'spkid': 704,
        'albedo': [0.23, 0.03, veverka],
        'H': [1.4, 0.2, veverka],
        'G': [0.10, 0.15, ''],
        'diameter': [1522.8, 5.2, 'Thomas et al. (1988). Icarus 73:427-441'],
        'GM': [192.4, 7.0, 'Jacobson et al. (2007). BAAS 39:453']
    },
    'miranda': {
        'spkid': 705,
        'albedo': [0.32, 0.03, veverka],
        'H': [3.4, 0.0, veverka],
        'G': [0.10, 0.15, ''],
        'diameter': [471.6, 1.4, 'Thomas et al. (1988). Icarus 73:427-441'],
        'GM': [4.4, 0.4, 'Jacobson et al. (2007). BAAS 39:453']
    },
    'caliban': {
        'spkid': 716,
        'albedo': [0.04, 0.0, 'Sheppard et al. (2005). AJ 129(1):518-525'],
        'H': [9.16, 0.04, 'Grav et al. (2004). AJ 613:L77–L80'],
        'G': [0.10, 0.15, ''],
        'diameter': [72, 0.0, 'Sheppard et al. (2005). AJ 129(1):518-525'],
        'rotation': [6.9190, 0.0082, 'Farkas-Takács et al. (2017) 154(3):119']
    },
    'sycorax': {
        'spkid': 717,
        'albedo': [0.04, 0.0, 'Sheppard et al. (2005). AJ 129(1):518-525'],
        'H': [7.50, 0.04, 'Grav et al. (2004). AJ 613:L77–L80'],
        'G': [0.10, 0.15, ''],
        'diameter': [150, 0.0, 'Sheppard et al. (2005). AJ 129(1):518-525'],
        'rotation': [9.948, 0.019, 'Farkas-Takács et al. (2017) 154(3):119']
    },
    'nereid': {
        'spkid': 802,
        'albedo': [0.155, 0.0, 'Thomas et al. (1991). JGR 96(19):253'],
        'H': [4.44, 0.01, 'Grav et al. (2004). AJ 613:L77–L80'],
        'G': [0.10, 0.15, ''],
        'diameter': [340, 50, 'Thomas et al. (1991). JGR 96(19):253']
    }
}


def search_satdb(name):
    sat = satellites.get(name.lower())
    if sat is None:
        raise ValueError('specified object was not found')
    return sat
# end of hard coded block.



# remove this block for v1.0
class _PositionDict(dict):
    """This is a modified Dictionary object to allow switching on/off of data points.
    It also avoids user to change data.
    """

    def __setitem__(self, key, value):
        """ Redefines how to set a value to a key in the dictionary.
        It only sets a value if the key starts with '_occ_'. Otherwise, it only allows for the user to provide
        'on' or 'off' which is passed only to change the 'on' keyword.
        """
        status = {'on': True, 'off': False}
        n = 0
        if key.startswith('_occ_'):
            super().__setitem__(key[5:], value)
            n = 1
        elif key in self.keys():
            n = 1
            if value not in status.keys():
                raise ValueError("Value must be 'on' or 'off' only.")
            if type(self[key]) == _PositionDict:
                for k in self[key].keys():
                    self[key][k] = value
            elif key == 'on':
                super().__setitem__('on', status[value])
                if status[value]:
                    self['enable']()
                else:
                    self['disable']()
        else:
            if value not in status.keys():
                raise ValueError("Value must be 'on' or 'off' only.")
            for key1 in self.keys():
                if type(self[key1]) == _PositionDict:
                    if key in self[key1].keys():
                        n = 1
                        self[key1][key] = value
        if n == 0:
            raise KeyError('Key "{}" does not exist'.format(key))

    def __str__(self):
        out = []
        for key in self.keys():
            if key not in ['enable', 'disable']:
                out.append('{}: {}'.format(key, self[key]))
        out = '\n' + '\n'.join(out)
        return out.replace('\n', '\n  ')

    def __repr__(self):
        out = []
        for key in self.keys():
            if key not in ['enable', 'disable']:
                out.append('\'{}\': {}'.format(key, self[key].__repr__()))
        out = ',\n'.join(out)
        return '{' + out.replace('\n', '\n  ') + '}'
# end of block removal



import warnings

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.time import Time

from sora.config import input_tests
from sora.config.decorators import deprecated_alias, deprecated_function
from .meta import MetaStar
from .utils import search_star, van_belle, kervella, spatial_motion, choice_star

warnings.simplefilter('always', UserWarning)

__all__ = ['Star']


class Star(MetaStar):
    """Defines a star.

    Parameters
    ----------
    catalogue : `str`
        The catalogue to download data. It can be ``'gaiadr2'`` or ``'gaiaedr3'``.

    code : `str`
        Gaia Source code for searching in VizieR.

    coord : `str`, `astropy.coordinates.SkyCoord`
        If code is not given, coord nust have the coordinates RA and DEC of the
        star to search in VizieR: ``'hh mm ss.ss +dd mm ss.ss'``.

    ra : `int`, `float`
        Right Ascension, in deg.

    dec : `int`, `float`
        Declination, in deg.

    parallax : `int`, `float`. default=0
        Parallax, in mas.

    pmra : `int`, `float`, default=0
        Proper Motion in RA*, in mas/year.

    pmdec : `int`, `float`, default=0
        Proper Motion in DEC, in mas/year.

    rad_vel : `int`, `float`, default=0
        Radial Velocity, in km/s.

    epoch : `str`, `astropy.time.Time`, default='J2000'
        Epoch of the coordinates.

    nomad : `bool`
        If True, it tries to download the magnitudes from NOMAD catalogue.

    bjones : `bool`, default=True
        If True, it uses de star distance from Bailer-Jones et al. (2018).

    verbose : `bool`, default=True
        If True, it prints the downloaded information

    local : `bool`, default=False
        If True, it uses the given coordinate in 'coord' as final coordinate.

    Note
    ----
    The user can give either 'coord' or 'ra' and 'dec', but not both.

    To download the coordinates from Gaia, "local" must be set as False
    and the ("code") or ("coord") or ("ra" and "dec") must be given.

    All values downloaded from Gaia will replace the ones given by the user.

    """

    @deprecated_alias(log='verbose')  # remove this line in v1.0
    def __init__(self, catalogue='gaiaedr3', **kwargs):

        self._attributes = {}
        self.mag = {}
        self.errors = {'RA': 0*u.mas, 'DEC': 0*u.mas, 'Plx': 0*u.mas, 'pmRA': 0*u.mas/u.year,
                       'pmDE': 0*u.mas/u.year, 'rad_vel': 0*u.km/u.year}
        allowed_kwargs = ['bjones', 'code', 'coord', 'dec', 'epoch', 'local', 'verbose', 'nomad', 'parallax', 'pmdec', 'pmra',
                          'ra', 'rad_vel']
        input_tests.check_kwargs(kwargs, allowed_kwargs=allowed_kwargs)
        allowed_catalogues = ['gaiadr2', 'gaiaedr3']
        if catalogue not in allowed_catalogues:
            raise ValueError('Catalogue {} is not one of the allowed catalogues {}'.format(catalogue, allowed_catalogues))
        self._catalogue = {'gaiadr2': 'Gaia-DR2', 'gaiaedr3': 'Gaia-EDR3'}[catalogue]
        self._verbose = kwargs.get('verbose', True)
        local = kwargs.get('local', False)
        self.bjones = False
        if 'code' in kwargs:
            self.code = kwargs['code']
        if 'coord' in kwargs:
            if 'ra' in kwargs or 'dec' in kwargs:
                raise ValueError("User must give 'coord' or 'ra' and 'dec', not both")
            coord = SkyCoord(kwargs['coord'], unit=('hourangle', 'deg'))
            self.ra = coord.ra
            self.dec = coord.dec
        if 'ra' in kwargs and 'dec' in kwargs:
            self.ra = kwargs.get('ra')
            self.dec = kwargs.get('dec')
        self.parallax = kwargs.get('parallax', 0.0)
        self.pmra = kwargs.get('pmra', 0.0)
        self.pmdec = kwargs.get('pmdec', 0.0)
        self.rad_vel = kwargs.get('rad_vel', 0.0)
        self.epoch = kwargs.get('epoch', 'J2000')
        if local:
            if 'RA' not in self._attributes or 'DEC' not in self._attributes:
                raise ValueError("User must give 'ra' and 'dec' for local coordinates")
        else:
            if not hasattr(self, 'code') and 'RA' not in self._attributes:
                raise ValueError("User must give gaia Source ID 'code' or coordinates for the online search")
            self.__searchgaia(catalog=catalogue)
        if kwargs.get('nomad', True):
            self.__getcolors()
        try:
            self.bjones = kwargs.get('bjones', False)
        except ValueError:
            pass

    def set_magnitude(self, **kwargs):
        """Sets the magnitudes of a star.

        Parameters
        ----------
        band=value : `str`
            The star magnitude for given band. The band name can be any string
            the user wants.

        Examples
        --------
        To set the stars magnitude in the band G:\n
        >>> set_magnitude(G=10)

        To set the star's magnitude in the band K:\n
        >>> set_magnitude(K=15)

        To set the star's magnitude in a customized band:\n
        >>> set_magnitude(newband=6)
        """
        for key in kwargs:
            mag = input_tests.test_attr(kwargs[key], float, key)
            if key in self.mag:
                warnings.warn('{0} mag already defined. {0}={1} will be replaced by {0}={2}'.format(
                    key, self.mag[key], mag))
            self.mag[key] = mag

    def set_diameter(self, diameter):
        """Sets an user diameter for the star, in mas.

        Parameters
        ----------
        diameter : `int`, `float`
            Sets the user diameter of the star, in mas.
        """
        self.diameter_user = diameter * u.mas
        if diameter < 0:
            warnings.warn("negative sizes are converted to positive.")
            self.diameter_user = np.absolute(self.diameter_user)

    def van_belle(self):
        """Determines the diameter of a star in mas using equations from van Belle (1999).

        See: Publi. Astron. Soc. Pacific 111, 1515-1523:.
        """
        return van_belle(self.mag.get('B'), self.mag.get('V'), self.mag.get('K'))

    def kervella(self):
        """Determines the diameter of a star in mas using equations from Kervella et. al (2004).

        See: A&A Vol.  426, No.  1:.
        """
        return kervella(self.mag.get('B'), self.mag.get('V'), self.mag.get('K'))

    @deprecated_alias(log='verbose')  # remove this line in v1.0
    def apparent_diameter(self, distance, mode='auto', band='V', star_type='sg', verbose=True):
        """Calculates the apparent diameter of the star at a given distance.

        Parameters
        ----------
        distance : `int`, `float`
            Object geocentric distance, in AU.

        mode : `str`, default='auto'
            The mode to calculate the apparent diameter.\n
            - ``'user'``: calculates using user given diameter.\n
            - ``'gaia'``: calculates using diameter obtained from Gaia.\n
            - ``'kervella'``: calculates using Kervella equations.\n
            - ``'van_belle'``: calculates using van Belle equations.\n
            - ``'auto'``: tries all the above methods until it is able to calculate diameter.\n
            The order of try is the same as shown above (user, Gaia, Kervella, Van Belle).

        band : `str`
            The band filter to calculate the diameter. If mode is `kervella`
            or `van_belle`, the filter must be given, ``'B'`` or ``'V'``.
            If mode `auto`, ``'V'`` is selected.

        star_type :`str`
            Type of star to calculate the diameter. If mode is `van_belle`,
            the star type must be given. If mode is `auto`, ``star_type='sg'``.\n
            Accepted types:\n
            - ``'sg'`` for 'Super Giant'.\n
            - ``'ms'`` for 'Main Sequence'.\n
            - ``'vs'`` for 'Variable Star'.

        verbose : `bool`
            If True, it prints the mode used by `auto`.
        """
        try:
            distance = distance.to(u.km)
        except:
            distance = distance * u.AU
        if distance < 0:
            warnings.warn("negative distances are converted to positive.")
            distance = np.absolute(distance)

        if mode in ['user', 'auto']:
            try:
                diam = distance*np.tan(self.diameter_user)
                if verbose:
                    print('Calculating apparent diameter from user defined diameter')
                return diam.to(u.km)
            except:
                pass

        if mode == 'user':
            raise ValueError('User diameter must be informed.')

        if mode in ['gaia', 'auto']:
            try:
                diam = distance*np.tan(self.diameter_gaia)
                if verbose:
                    text = ''
                    if self.bjones:
                        text += ' + Bailer-Jones et al. (2018)'
                    print('Apparent diameter using Gaia' + text)
                return diam.to(u.km)
            except:
                pass

        if mode == 'gaia':
            raise ValueError('It is not possible to calculate star diameter from Gaia.')

        if band not in ['B', 'V']:
            raise KeyError('band must be informed as "B", or "V"')

        if mode in ['kervella', 'auto']:
            diam_kerv = self.kervella().get(band)
            if diam_kerv is None:
                raise ValueError('Diameter could not be calculated for given band')
            if verbose:
                print('Apparent diameter using Kervella et al. (2004)')
            diam = distance*np.tan(diam_kerv)
            return diam.to(u.km)

        if star_type not in ['sg', 'ms', 'vs']:
            raise KeyError('star_type must be informed as "sg", "ms" or "vs"')

        if mode in ['van_belle', 'auto']:
            diam_van = self.van_belle().get(star_type)
            if diam_van is None:
                raise ValueError('Diameter could not be calculated using Van Belle')
            diam_van = diam_van.get(band)
            if diam_van is None:
                raise ValueError('Diameter could not be calculated for given band')
            if verbose:
                print('Apparent diameter using van Belle (1999)')
            diam = distance*np.tan(diam_van)
            return diam.to(u.km)

        raise AttributeError("Star apparent diameter could not be calculated. ",
                             "Please define star diameter or B,V,K magnitudes.")

    def __searchgaia(self, catalog):
        """Searches for the star position in the Gaia catalogue and save information.

        Parameters
        ----------
        catalog : `str`
            The catalogue to download data. It can be ``'gaiadr2'`` or ``'gaiaedr3'``.
        """
        catalogues = {'gaiadr2': 'I/345/gaia2', 'gaiaedr3': 'I/350/gaiaedr3'}
        cat = catalogues[catalog]
        if hasattr(self, 'code'):
            catalogue = search_star(code=self.code, columns=['**'], catalog=cat, verbose=self._verbose)
        else:
            catalogue = search_star(coord=self.coord, columns=['**'], radius=2*u.arcsec,
                                    catalog=cat, verbose=self._verbose)
        if len(catalogue) == 0:
            raise ValueError('No star was found. It does not exist or VizieR is out.')
        catalogue = catalogue[0]
        if len(catalogue) > 1:
            if self._verbose:
                print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
                print('The list below is sorted by distance. Please select the correct star')
            catalogue = choice_star(catalogue, self.coord, ['RA_ICRS', 'DE_ICRS', 'Gmag'], source='gaia')
        self.code = catalogue['Source'][0]
        self.ra = catalogue['RA_ICRS'][0]*u.deg
        self.dec = catalogue['DE_ICRS'][0]*u.deg
        self.pmra = catalogue['pmRA'][0]*u.mas/u.year
        self.pmdec = catalogue['pmDE'][0]*u.mas/u.year
        self.epoch = Time(catalogue['Epoch'][0], format='jyear')
        self.parallax = catalogue['Plx'][0]*u.mas
        rv_name = {'gaiadr2': 'RV', 'gaiaedr3': 'RVDR2'}
        self.rad_vel = catalogue[rv_name[catalog]][0]*u.km/u.s
        self.set_magnitude(G=catalogue['Gmag'][0])

        self.meta_gaia = {c: catalogue[c][0] for c in catalogue.columns}

        self.errors['RA'] = self.meta_gaia['e_RA_ICRS']*u.mas
        self.errors['DEC'] = self.meta_gaia['e_DE_ICRS']*u.mas
        self.errors['Plx'] = self.meta_gaia['e_Plx']*u.mas
        self.errors['pmRA'] = self.meta_gaia['e_pmRA']*(u.mas/u.yr)
        self.errors['pmDE'] = self.meta_gaia['e_pmDE']*(u.mas/u.yr)
        erv_name = {'gaiadr2': 'e_RV', 'gaiaedr3': 'e_RVDR2'}
        self.errors['rad_vel'] = self.meta_gaia[erv_name[catalog]]*(u.km/u.s)

        A = (1*u.AU).to(u.km).value
        cov = np.zeros((6, 6))
        a = ['RA', 'DE', 'Plx', 'pmRA', 'pmDE']
        for i in np.arange(5):
            v1 = 'e_' + a[i]
            if i in [0, 1]:
                v1 += '_ICRS'
            for j in np.arange(i, 5):
                v2 = 'e_' + a[j]
                if j in [0, 1]:
                    v2 += '_ICRS'
                if i == j:
                    x = self.meta_gaia[v1]**2
                    if not np.ma.core.is_masked(x):
                        cov[i, i] = x
                else:
                    x = self.meta_gaia[a[i]+a[j]+'cor']*self.meta_gaia[v1]*self.meta_gaia[v2]
                    if not np.ma.core.is_masked(x):
                        cov[i, j] = x
                        cov[j, i] = cov[i, j]
            x = cov[i, 2]*(self.meta_gaia[rv_name[catalog]]/A)
            if not np.ma.core.is_masked(x):
                cov[i, 5] = x
                cov[5, i] = cov[i, 5]
        x = cov[2, 2]*(self.meta_gaia[rv_name[catalog]]**2 + self.meta_gaia[erv_name[catalog]]**2)/(A**2) \
            + (self.meta_gaia['Plx']*self.meta_gaia[erv_name[catalog]]/A)**2
        if not np.ma.core.is_masked(x):
            cov[5, 5] = x
        cov[np.where(np.isnan(cov))] = 0.0
        self.cov = cov

        if self._verbose:
            print('1 {} star found G={}'.format(self._catalogue, catalogue['Gmag'][0]))
            print('star coordinate at J{}: RA={} +/- {}, DEC={} +/- {}'.format(self.epoch.jyear,
                  self.ra.to_string(u.hourangle, sep='hms', precision=5), self.errors['RA'],
                  self.dec.to_string(u.deg, sep='dms', precision=4), self.errors['DEC']))

    def __getcolors(self):
        """ Searches for the B,V,K magnitudes of the star in the NOMAD catalogue on VizieR.
        """
        columns = ['RAJ2000', 'DEJ2000', 'Bmag', 'Vmag', 'Rmag', 'Jmag', 'Hmag', 'Kmag']
        catalogue = search_star(coord=self.coord, columns=columns, radius=2*u.arcsec,
                                catalog='I/297/out', verbose=self._verbose)
        if len(catalogue) == 0:
            if self._verbose:
                warnings.warn('No star was found on NOMAD that matches the star')
            return
        catalogue = catalogue[0]
        if len(catalogue) > 1:
            print('{} stars were found within 2 arcsec from given coordinate.'.format(len(catalogue)))
            print('The list below is sorted by distance. Please select the correct star')
            if hasattr(self.mag, 'G'):
                print('Star G mag: {}'.format(self.mag['G']))
            catalogue = choice_star(catalogue, self.coord, ['RAJ2000', 'DEJ2000', 'Bmag', 'Vmag',
                                                            'Rmag', 'Jmag', 'Hmag', 'Kmag'], source='nomad')
            if catalogue is None:
                return
        errors = []
        for mag in ['B', 'V', 'R', 'J', 'H', 'K']:
            name = mag + 'mag'
            if np.ma.core.is_masked(catalogue[name][0]):
                errors.append(mag)
                continue
            self.set_magnitude(**{mag: catalogue[name][0]})
        if len(errors) > 0 and self._verbose:
            print('Magnitudes in {} were not located in NOMAD'.format(errors))

    # remove this block for v1.0
    @deprecated_function(message="Please use get_position(time=time, observer='geocenter')")
    def geocentric(self, time):
        """ Calculates the position of the star, propagating the position using parallax and proper motion

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to apply proper motion and calculate parallax. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.
        """
        return self.get_position(time=time, observer='geocenter')

    @deprecated_function(message="Please use get_position(time=time, observer='barycenter')")
    def barycentric(self, time):
        """Calculates the position of the star using proper motion.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to apply proper motion. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.
        """
        return self.get_position(time=time, observer='barycenter')
    # end of block removal

    def get_position(self, time, observer='geocenter'):
        """Calculates the position of the star for given observer,
        propagating the position using parallax and proper motion

        Parameters
        ----------
        time : `float`, `astropy.time.Time`
            Reference time to apply proper motion and calculate parallax. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        observer : `str`, `sora.observer.Observer`, `sora.observer.Spacecraft`
            Observer of the star t calculate position. It can be 'geocenter' for a geocentric
            coordinate, 'barycenter' for a barycenter coordinate, or a sora observer object.

        Returns
        -------
        coord : `astropy.coordinates.SkyCoord`
            Astropy SkyCoord object with the star coordinates at the given time.
        """
        from astropy.coordinates import SphericalRepresentation, SkyOffsetFrame, ICRS
        from sora import Observer, Spacecraft

        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        if observer not in ['geocenter', 'barycenter'] and not isinstance(observer, (Observer, Spacecraft)):
            raise ValueError("'observer' must be an Observer object or one of the following"
                             " strings: ['geocenter', 'barycenter]")

        def apply_offset(coord):
            if not hasattr(self, 'offset'):
                return coord
            star_frame = SkyOffsetFrame(origin=coord)
            new_pos = SkyCoord(lon=self.offset.d_lon_coslat, lat=self.offset.d_lat, frame=star_frame)
            p = new_pos.transform_to(ICRS)
            return SkyCoord(ra=p.ra, dec=p.dec, distance=p.distance)

        dt = time - self.epoch
        if not time.isscalar:
            if time.max() - time.min() > 1*u.day:
                raise ValueError('list of times must be in a interval of 1 day to process.')
            dt = dt[0]
        bar_star = spatial_motion(self.ra, self.dec, self.pmra, self.pmdec, self.parallax, self.rad_vel, dt=dt.jd)

        if observer == "barycenter" or self.coord.distance.unit.is_unity() or np.isnan(self.coord.distance):
            return apply_offset(bar_star)

        if observer == "geocenter":
            observer = Observer(code='500', ephem='horizons')

        bar_obs = observer.get_vector(time=time, origin='barycenter')

        topo = bar_star.cartesian - bar_obs.cartesian
        topo = topo.represent_as(SphericalRepresentation)
        topo = SkyCoord(topo.lon, topo.lat, topo.distance)
        return apply_offset(topo)

    def error_at(self, time):
        """Estimates the star position error at a given time.

        Parameters
        ----------
        time : `str`, `astropy.time.Time`
            Reference time to project star error. It can be a string
            in the ISO format (yyyy-mm-dd hh:mm:ss.s) or an astropy Time object.

        Returns
        -------
        errors : `list`
            In RA* and DEC.

        """
        try:
            time = Time(time)
        except:
            time = Time(time, format='jd', scale='utc')
        dt = time - self.epoch
        n_coord, errors = spatial_motion(self.ra, self.dec, self.pmra, self.pmdec, self.parallax,
                                         self.rad_vel, dt=dt.jd, cov_matrix=self.cov)
        return errors[0]*u.mas, errors[1]*u.mas

    def add_offset(self, da_cosdec, ddec):
        """Adds an offset to the star position.

        Parameters
        ----------
        da_cosdec : `int`, `float`
            Offset in Delta_alpha_cos_delta, in mas.

        ddec : `int`, `float`
            Offset in Delta_delta, in mas.
        """
        from astropy.coordinates import SphericalCosLatDifferential

        dadc = input_tests.test_attr(da_cosdec, float, 'da_cosdec')
        dd = input_tests.test_attr(ddec, float, 'ddec')
        self.offset = SphericalCosLatDifferential(dadc * u.mas, dd * u.mas, 0.0 * u.km)

    def __str__(self):
        """String representation of the Star class.
        """
        out = ''
        if hasattr(self, 'code'):
            out += '{} star Source ID: {}\n'.format(self._catalogue, self.code)
        else:
            out += 'User coordinates\n'
        out += ('ICRS star coordinate at J{}:\n'
                'RA={} +/- {:.4f}, DEC={} +/- {:.4f}\n'
                'pmRA={:.3f} +/- {:.3f} mas/yr, pmDEC={:.3f} +/- {:.3f} mas/yr\n'
                'Plx={:.4f} +/- {:.4f} mas, Rad. Vel.={:.2f} +/- {:.2f} km/s \n\n'.format(
                    self.epoch.jyear, self.ra.to_string(u.hourangle, sep='hms', precision=5),
                    self.errors['RA'], self.dec.to_string(u.deg, sep='dms', precision=4), self.errors['DEC'],
                    self.pmra.value, self.errors['pmRA'].value, self.pmdec.value, self.errors['pmDE'].value,
                    self.parallax.value, self.errors['Plx'].value, self.rad_vel.value, self.errors['rad_vel'].value))
        if hasattr(self, 'offset'):
            out += 'Offset Apllied: d_alpha_cos_dec = {}, d_dec = {}\n'.format(
                self.offset.d_lon_coslat, self.offset.d_lat)
        out += 'Magnitudes:'
        mag_out = [' {}: {:6.3f}'.format(mag, self.mag[mag]) for mag in self.mag]
        out_mag = []
        for i, mag in enumerate(mag_out):
            if i % 6 == 0:
                out_mag.append([])
            out_mag[-1].append(mag)
        out += (',\n'+' '*11).join([','.join(out_i) for out_i in out_mag])
        out += '\n\n'
        if self.diameter_gaia is not None:
            text = ''
            if self.bjones:
                text += ' + Bailer-Jones et al. (2018)'
            out += 'Apparent diameter: {:.4f}, Source: Gaia-DR2{}\n'.format(self.diameter_gaia, text)
        if hasattr(self, 'diameter_user'):
            out += 'Apparent diameter: {:.4f}, Source: User\n'.format(self.diameter_user)
        kerv = self.kervella()
        if kerv:
            out += 'Apparent diameter from Kervella et. al (2004):\n'
            out += '   ' + ','.join([' {}: {:.4f}'.format(k, v) for k, v in kerv.items()])
        vanb = self.van_belle()
        if vanb:
            out += '\nApparent diameter from van Belle (1999):'
            for key, value in vanb.items():
                out += '\n    {}:'.format(key)
                out += ','.join([' {}: {:.4f}'.format(k, v) for k, v in value.items()])
        return out
